import numpy as np

import xtrack as xt

x_co = [1e-3, 2e-3]
px_co = [2e-6, -3e-6]
y_co = [3e-3, 4e-3]
py_co = [4e-6, -5e-6]
betx = [1., 2.]
bety = [3., 4.]
alfx = [0, 0.1]
alfy = [0.2, 0.]
dx = [10, 0]
dy = [0, 20]
dpx = [0.7, -0.3]
dpy = [0.4, -0.6]
bets = 1e-3
dnqx = [0.21, 2, 30, 400]
dnqy = [0.32, 3, 40, 500]

segm_1 = xt.LineSegmentMap(
        qx=0.4, qy=0.3, qs=0.0001,
        bets=bets, length=0.1,
        betx=[betx[0], betx[1]],
        bety=[bety[0], bety[1]],
        alfx=[alfx[0], alfx[1]],
        alfy=[alfy[0], alfy[1]],
        dx=[dx[0], dx[1]],
        dpx=[dpx[0], dpx[1]],
        dy=[dy[0], dy[1]],
        dpy=[dpy[0], dpy[1]],
        x_ref=[x_co[0], x_co[1]],
        px_ref=[px_co[0], px_co[1]],
        y_ref=[y_co[0], y_co[1]],
        py_ref=[py_co[0], py_co[1]])
segm_2 = xt.LineSegmentMap(
        qs=0.0003,
        bets=bets, length=0.2,
        dnqx=dnqx, dnqy=dnqy,
        betx=[betx[1], betx[0]],
        bety=[bety[1], bety[0]],
        alfx=[alfx[1], alfx[0]],
        alfy=[alfy[1], alfy[0]],
        dx=[dx[1], dx[0]],
        dpx=[dpx[1], dpx[0]],
        dy=[dy[1], dy[0]],
        dpy=[dpy[1], dpy[0]],
        x_ref=[x_co[1], x_co[0]],
        px_ref=[px_co[1], px_co[0]],
        y_ref=[y_co[1], y_co[0]],
        py_ref=[py_co[1], py_co[0]])

line = xt.Line(elements=[segm_1, segm_2], particle_ref=xt.Particles(p0c=1e9))
line.build_tracker()

tw4d = line.twiss(method='4d')
tw6d = line.twiss()

assert np.isclose(tw6d.qs, 0.0004, atol=1e-7, rtol=0)
assert np.isclose(tw6d.bets0, 1e-3, atol=1e-7, rtol=0)

for tw in [tw4d, tw6d]:

    assert np.isclose(tw.qx, 0.4 + 0.21, atol=1e-7, rtol=0)
    assert np.isclose(tw.qy, 0.3 + 0.32, atol=1e-7, rtol=0)

    assert np.isclose(tw.dqx, 2, atol=1e-5, rtol=0)
    assert np.isclose(tw.dqy, 3, atol=1e-5, rtol=0)

    assert np.allclose(tw.s, [0, 0.1, 0.1 + 0.2], atol=1e-7, rtol=0)
    assert np.allclose(tw.mux, [0, 0.4, 0.4 + 0.21], atol=1e-7, rtol=0)
    assert np.allclose(tw.muy, [0, 0.3, 0.3 + 0.32], atol=1e-7, rtol=0)

    assert np.allclose(tw.betx, [1, 2, 1], atol=1e-7, rtol=0)
    assert np.allclose(tw.bety, [3, 4, 3], atol=1e-7, rtol=0)

    assert np.allclose(tw.alfx, [0, 0.1, 0], atol=1e-7, rtol=0)
    assert np.allclose(tw.alfy, [0.2, 0, 0.2], atol=1e-7, rtol=0)

    assert np.allclose(tw.dx, [10, 0, 10], atol=1e-4, rtol=0)
    assert np.allclose(tw.dy, [0, 20, 0], atol=1e-4, rtol=0)
    assert np.allclose(tw.dpx, [0.7, -0.3, 0.7], atol=1e-5, rtol=0)
    assert np.allclose(tw.dpy, [0.4, -0.6, 0.4], atol=1e-5, rtol=0)

    assert np.allclose(tw.x, [1e-3, 2e-3, 1e-3], atol=1e-7, rtol=0)
    assert np.allclose(tw.px, [2e-6, -3e-6, 2e-6], atol=1e-12, rtol=0)
    assert np.allclose(tw.y, [3e-3, 4e-3, 3e-3], atol=1e-7, rtol=0)
    assert np.allclose(tw.py, [4e-6, -5e-6, 4e-6], atol=1e-12, rtol=0)


    assert np.allclose(tw.py, [4e-6, -5e-6, 4e-6], atol=1e-12, rtol=0)

    chroma_table = line.get_non_linear_chromaticity((-1e-2, 1e-2),
                                                    num_delta=25)
    assert np.allclose(chroma_table.dnqx[1:], dnqx[1:], atol=1e-5, rtol=0)
    assert np.allclose(chroma_table.dnqy[1:], dnqy[1:], atol=1e-5, rtol=0)



import xtrack as xt

env = xt.Environment()

env.call('pimm_lattice.py')
env.call('pimm_optics.py')

env['pimm'].particle_ref = xt.Particles(q0=1, mass0=xt.PROTON_MASS_EV,
                                        kinetic_energy0=200e6) # eV

env['pimm'].twiss4d().plot()


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

from .elements import *
from .exciter import Exciter
from .apertures import *
from .magnets import Magnet
from .beam_interaction import BeamInteraction, ParticlesInjectionSample
from .slice_elements import (ThinSliceQuadrupole, ThinSliceSextupole,
                             ThinSliceOctupole, ThinSliceBend,
                             ThinSliceRBend,
                             ThinSliceBendEntry, ThinSliceBendExit,
                             ThinSliceRBendEntry, ThinSliceRBendExit,
                             ThinSliceQuadrupoleEntry, ThinSliceQuadrupoleExit,
                             ThinSliceSextupoleEntry, ThinSliceSextupoleExit,
                             ThinSliceOctupoleEntry, ThinSliceOctupoleExit)
from .slice_elements_thick import (ThickSliceBend, ThickSliceRBend,
                                   ThickSliceQuadrupole, ThickSliceSextupole,
                                   ThickSliceOctupole, ThickSliceSolenoid,
                                   DriftSliceOctupole, DriftSliceSextupole,
                                   DriftSliceQuadrupole, DriftSliceBend,
                                   DriftSliceRBend, DriftSlice)
from .rft_element import RFT_Element
from ..base_element import BeamElement

element_classes = tuple(v for v in globals().values() if isinstance(v, type) and issubclass(v, BeamElement))



import bpmeth
import xtrack as xt
import sympy

import numpy as np

b1 = "0.12"
h = "0.1"
length = 0.5

s = sympy.symbols('s')
b1_f = sympy.lambdify(s, b1, modules='numpy')

# Vector potential (bs, b1, a1, b2, a2, b3, a3, ,... as function of s)
A_magnet = bpmeth.GeneralVectorPotential(hs=h, b=(b1,))

# Make Hamiltonian for a defined length
H_magnet = bpmeth.Hamiltonian(length=length, curv=float(h), vectp=A_magnet)

# This guy is able to track an Xsuite particle!
p0 = xt.Particles(x=np.linspace(-1e-3, 1e-3, 10), energy0=10e9, mass0=xt.ELECTRON_MASS_EV)

p_silke = p0.copy()
sol = H_magnet.track(p_silke, return_sol=True)

p_xsuite = p0.copy()
bb = xt.Bend(h=float(h), length=length, k0=float(b1))

bb.model =  'bend-kick-bend'
bb.track(p_xsuite)

# I want to see the trajectory along the object
import matplotlib.pyplot as plt
plt.close('all')
plt.figure()
for ss in sol:
    plt.plot(ss.t, ss.y[0])
plt.xlabel('s [m]')
plt.ylabel('x [m]')

plt.show()



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
import matplotlib.patches
import matplotlib.collections
import matplotlib.pyplot as plt
#plt.style.use('kostas')

def initial_xy_polar(r_min, r_max, r_N, theta_min, theta_max, theta_N):
    return np.array([[(r*np.cos(theta),r*np.sin(theta)) for r in np.linspace(r_min,r_max,r_N)] for theta in np.linspace(theta_min,theta_max,theta_N)])

def initial_xy_cartesian(x_min, x_max, x_N, y_min, y_max, y_N):
    return np.array([[(x,y) for x in np.linspace(x_min,x_max,x_N)] for y in np.linspace(y_min,y_max,y_N)])


def draw_footprint(A, axis_object=None, figure_object=None, axis=0, linewidth=4):
    '''
    Input A should be a 3-D numpy array with shape (Nx,Ny,2)
    representing a 2-D array of (x,y) points. This function
    will draw lines between adjacent points in the 2-D array.
    '''
    if len(A.shape) != 3:
        print('ERROR: Invalid input matrix')
        return None
    if A.shape[2] != 2:
        print('ERROR: Points are not defined in 2D space')
        return None

    sx = A.shape[0]-1
    sy = A.shape[1]-1

    p1 = A[:-1,:-1,:].reshape(sx*sy,2)[:,:]
    p2 = A[1:,:-1,:].reshape(sx*sy,2)[:]
    p3 = A[1:,1:,:].reshape(sx*sy,2)[:]
    p4 = A[:-1,1:,:].reshape(sx*sy,2)[:]

    #Stack endpoints to form polygons
    Polygons = np.stack((p1,p2,p3,p4))
    #transpose polygons
    Polygons = np.transpose(Polygons,(1,0,2))
    patches = list(map(matplotlib.patches.Polygon,Polygons))

    #assign colors
    patch_colors = [(0,0,0) for a in Polygons]
    patch_colors[(sx-1)*sy:] = [(0,1,0)]*sy
    patch_colors[(sy-1)::sy] = [(0,0,1)]*sx

    p_collection = matplotlib.collections.PatchCollection(patches,facecolors=[],linewidth=linewidth,edgecolor=patch_colors)

    if axis_object is None:
        if figure_object:
            fig = figure_object
        else:
            fig = plt.figure()
        if len(fig.axes) == 0:
            plt.subplot(1,1,1)
        if axis >= len(fig.axes) or axis < 0:
            i = 0
        else:
            i = axis
        ax = fig.axes[i]
    else:
        ax = axis_object
        fig = None

    ax.add_collection(p_collection)

    return fig

def example():
    coords_matrix = initial_xy_polar(0.01,5,11,0.,np.pi/2.,10)
    fig = draw_footprint(coords_matrix)
    fig.axes[0].set_xlim(-0.05,5.1)
    fig.axes[0].set_ylim(-0.05,5.1)



import xtrack as xt
import xpart as xp
from xpart.longitudinal.rf_bucket import RFBucket

import numpy as np
from scipy.constants import c as clight
from scipy.constants import e as qe
import matplotlib.pyplot as plt

gamma0 = 5 # defines the energy of the beam
gamma_transition = 4.
momentum_compaction_factor = 1 / gamma_transition**2
compensate_phase = True

particle_ref = xt.Particles(gamma0=gamma0,
                            mass0=xt.PROTON_MASS_EV)

circumference = 1000.
t_rev = circumference / (particle_ref.beta0[0] * clight)
f_rev = 1 / t_rev

energy_ref_increment =  50e3 # eV

eta = momentum_compaction_factor - 1 / particle_ref.gamma0[0]**2

h_rf = 40

f_rf = h_rf * f_rev
v_rf = 100e3
lag_rf = 180. if eta > 0. else 0.

# Compute momentum increment using auxiliary particle
dp0c_eV = energy_ref_increment / particle_ref.beta0[0]

if compensate_phase:
    phi = np.arcsin(dp0c_eV * particle_ref.beta0[0] / v_rf)
    if eta > 0:
        phi = np.pi - phi
    lag_rf = np.rad2deg(phi)

otm = xt.LineSegmentMap(
    betx=1., bety=1,
    qx=6.3, qy=6.4,
    momentum_compaction_factor=momentum_compaction_factor,
    longitudinal_mode="nonlinear",
    voltage_rf=v_rf,
    frequency_rf=f_rf,
    lag_rf=lag_rf,
    length=circumference,
    energy_ref_increment=energy_ref_increment
)

line = xt.Line(elements=[otm], particle_ref=particle_ref)

tw = line.twiss()

delta_test = np.linspace(0, 5e-3, 20)
p = line.build_particles(delta=delta_test)

line.track(p, turn_by_turn_monitor=True, num_turns=1000)
mon = line.record_last_track

p_gauss, matcher = xp.generate_matched_gaussian_bunch(
    line=line,
    num_particles=10_000,
    nemitt_x=2.5e-6,
    nemitt_y=2.5e-6,
    sigma_z=2.,
    return_matcher=True)

rfb = line._get_bucket()
z_separatrix = np.linspace(rfb.z_left, rfb.z_right, 1000)
delta_separatrix = rfb.separatrix(z_separatrix)

plt.close('all')
plt.figure(1)
plt.plot(p_gauss.zeta, p_gauss.delta, '.', color='k', alpha=0.5)
plt.plot(mon.zeta.T, mon.delta.T, color='C0')
plt.plot(z_separatrix, delta_separatrix, color='C1')
plt.plot(z_separatrix, -delta_separatrix, color='C1')
plt.xlabel('zeta [m]')
plt.ylabel('delta')

# Check the force and hamiltonian
z_test = np.linspace(-30, 30, 1000)
force = rfb.total_force(z_test)
hamiltonian = rfb.hamiltonian(z_test, 0)

plt.figure(2)
ax1 = plt.subplot(2,1,1)
ax1.plot(z_test, force)
ax1.set_ylabel(r'F($\zeta$)')
ax2 = plt.subplot(2,1,2, sharex=ax1)
ax2.plot(z_test, hamiltonian)
ax2.set_xlabel(r'$\zeta$ [m]')
ax2.set_ylabel(r'$H(\zeta, 0)$')

# Check hamiltonian on the delta axis
delta_test = np.linspace(-1e-2, 1e-2, 1000)
plt.figure(3)
plt.plot(delta_test, matcher.rfbucket.hamiltonian(0, delta_test))
plt.plot(delta_test, matcher.psi_object.H(0, delta_test))
plt.plot(delta_test, matcher.rfbucket.hamiltonian(0, delta_test, make_convex=True), '--')
plt.xlabel(r'$\delta$')
plt.ylabel(r'$H(0, \delta)$')

plt.show()


import itertools
import pathlib

import numpy as np
from cpymad.madx import Madx
from scipy.constants import c as clight
from scipy.special import factorial

import xobjects as xo
import xpart as xp
import xtrack as xt
import xtrack.mad_loader
from xtrack import MadLoader

test_data_folder = pathlib.Path(
    __file__).parent.joinpath('../test_data').absolute()


def test_non_zero_index():
    lst = [1, 2, 3, 0, 0, 0]
    assert xtrack.mad_loader.non_zero_len(lst) == 3


def test_add_lists():
    a = [1, 2, 3, 1, 1, 1]
    b = [1, 1, 1, 4, 5, 6]
    c = xtrack.mad_loader.add_lists(a, b, 8)
    assert c == [2, 3, 4, 5, 6, 7, 0, 0]


def test_add_lists():
    a = [1, 2, 3, 1, 1, 1]
    b = [1, 1, 1, 4, 5, 6, 7, 8]
    c = xtrack.mad_loader.add_lists(a, b, 8)
    assert c == [2, 3, 4, 5, 6, 7, 7, 8]

    a = [1, 2, 3, 1, 1, 1]
    b = [1, 1, 1, 4, 5, 6, 7, 8]
    c = xtrack.mad_loader.add_lists(a, b, 10)
    assert c == [2, 3, 4, 5, 6, 7, 7, 8, 0, 0]


def test_tilt_shift_and_errors():
    mad = Madx(stdout=False)

    src = """
    k1=0.2;
    tilt=0.1;

    elm: multipole,
            knl:={0.1,-k1,0.3},
            ksl={0.,0.2,-0.3,4},
            angle=0.1,
            tilt=0.2,
            lrad=1,
            apertype="rectellipse",
            aperture={0.1,0.2,0.11,0.22},
            aper_tol={0.1,0.2,0.3},
            aper_tilt:=tilt,
            aper_offset={0.2, 0.3};

    seq: sequence, l=1;
    elm1:elm, at=0;
    mk:marker, at=0.1;
    mk2:marker, at=0.1,aperture={0.1,0.2,0.11,0.22},apertype="rectellipse";
    elm2:elm, at=0.5;
    elm3:elm, at=0.5;
    endsequence;


    beam;
    use,sequence=seq;

    select,pattern=elm,flag=error;
    efcomp,order=0,radius=0.3,
    dkn={0.01,-0.02,0.03,0.2},
    dks={-0.01,0.02,-0.03,0.3,5},
    dknr={0.02,-0.04,0.06},
    dksr={-0.03,0.05,-0.07};

    ealign,
    DX=0.1, DY=0.3, DS=0.0,
    DPHI=0.0, DTHETA=0.0, DPSI=0.3,
    AREX=0.2, AREY=0.3;

    twiss,betx=1,bety=1,x=0.1,y=0.2,t=0.3;

    track,onepass;
    start,x=0.1,y=0.2,t=0.3;
    run,turns=1;
    endtrack;
    value,table(twiss,elm,x);
    value,table(tracksumm,x,2);
    value,table(twiss,elm,px);
    value,table(tracksumm,px,2);
    value,table(twiss,elm,y);
    value,table(tracksumm,y,2);
    value,table(twiss,elm,py);
    value,table(tracksumm,py,2);
    """

    mad.input(src)

    from xtrack import MadLoader

    def gen_options(opt1):
        for v in itertools.product(*([[True, False]] * len(opt1))):
            yield dict(zip(opt1, v))

    opt1 = ["enable_expressions",
        "enable_errors",
        "enable_apertures"]

    opt2 = ["skip_markers", "merge_drifts", "merge_multipoles"]

    for opt in gen_options(opt1):
        ml = MadLoader(mad.sequence.seq, **opt)
        line = ml.make_line()

        if opt['enable_apertures'] and opt['enable_errors']:
            assert np.all(np.array(line.element_names) == (
                'seq$start',
                'elm1_aper',
                'elm1',
                'drift_0', 'mk', 'mk2_aper', 'mk2', 'drift_1',
                'elm2_aper',
                'elm2',
                'elm3_aper',
                'elm3',
                'drift_2', 'seq$end'))

        elif opt['enable_apertures'] and not (opt['enable_errors']):
            assert np.all(np.array(line.element_names) == (
                'seq$start',
                'elm1_aper',
                'elm1',
                'drift_0',
                'mk',
                'mk2_aper', 'mk2', 'drift_1',
                'elm2_aper',
                'elm2',
                'elm3_aper',
                'elm3', 'drift_2', 'seq$end'
            ))
        elif not (opt['enable_apertures']) and opt['enable_errors']:
            assert np.all(np.array(line.element_names) == (
                'seq$start', 'elm1',
                'drift_0', 'mk', 'mk2',
                'drift_1', 'elm2',
                'elm3',
                'drift_2', 'seq$end'))
        elif not (opt['enable_apertures']) and not (opt['enable_errors']):
            assert np.all(np.array(line.element_names) == (
                'seq$start', 'elm1',
                'drift_0', 'mk', 'mk2', 'drift_1', 'elm2',
                'elm3',
                'drift_2', 'seq$end'))

        for nn in line.element_names:
            if 'tilt' in nn:
                assert isinstance(line[nn], xt.SRotation)
            elif 'offset' in nn:
                assert isinstance(line[nn], xt.XYShift)

        mad_elm2 = mad.sequence['seq'].expanded_elements['elm2']

        on_err = int(opt['enable_errors'])
        if opt['enable_apertures']:
            xo.assert_allclose(line['elm2_aper'].rot_s_rad,
                               mad_elm2.aper_tilt,
                               rtol=0, atol=1e-13)

            xo.assert_allclose(line['elm2_aper'].shift_x,
                               on_err * mad_elm2.align_errors.arex + mad_elm2.aper_offset[0],
                               rtol=0, atol=1e-13)
            xo.assert_allclose(line['elm2_aper'].shift_y,
                               on_err * mad_elm2.align_errors.arey + mad_elm2.aper_offset[1],
                               rtol=0, atol=1e-13)

            assert isinstance(line['elm2_aper'], xt.LimitRectEllipse)
            assert line['elm2_aper'].max_x == .1
            assert line['elm2_aper'].max_y == .2
            assert line['elm2_aper'].a_squ == .11 ** 2
            assert line['elm2_aper'].b_squ == .22 ** 2

        xo.assert_allclose(line['elm2'].rot_s_rad,
                           (mad_elm2.tilt + on_err * mad_elm2.align_errors.dpsi),
                           rtol=0, atol=1e-13)

        if opt['enable_errors']:
            xo.assert_allclose(line['elm2'].shift_x,
                               on_err * mad_elm2.align_errors.dx, rtol=0, atol=1e-13)
            xo.assert_allclose(line['elm2'].shift_y,
                               on_err * mad_elm2.align_errors.dy, rtol=0, atol=1e-13)

        for ii in range(line['elm2'].order + 1):
            ref = 0
            if len(mad_elm2.knl) > ii:
                ref += mad_elm2.knl[ii]
            if opt['enable_errors'] and len(mad_elm2.field_errors.dkn) > ii:
                ref += mad_elm2.field_errors.dkn[ii]
            xo.assert_allclose(line['elm2'].knl[ii], ref, rtol=0, atol=1e-13)

        for ii in range(line['elm2'].order + 1):
            ref = 0
            if len(mad_elm2.ksl) > ii:
                ref += mad_elm2.ksl[ii]
            if opt['enable_errors'] and len(mad_elm2.field_errors.dks) > ii:
                ref += mad_elm2.field_errors.dks[ii]
            xo.assert_allclose(line['elm2'].ksl[ii], ref, rtol=0, atol=1e-13)

    for opt in gen_options(opt2):
        ml = MadLoader(mad.sequence.seq, **opt)
        line = list(ml.iter_elements())


def test_thick_errors():
    mad = Madx()

    mad.input('''
        k0 = 0.1;
        k1 = 0.2;
        k2 = 0.3;
        k3 = 0.4;
        ks = 0.5;
        l = 5.;
        radius = 0.3;

        dknr0 = 0.02;
        dknr1 = -0.04;
        dknr2 = 0.06;
        dkns3 = 0.2;
        dksr0 = 0.03;
        dksr1 = -0.05;
        dksr2 = 0.07;
        dksr3 = 0.3;

        bend: sbend, k0 := k0, l = 5;
        quad: quadrupole, k1 := k1, l = 5.;
        sext: sextupole, k2 := k2, l = 5.;
        octu: octupole, k3 := k3, l = 5.;
        sole: solenoid, ks := ks, l = 5.;

        seq: sequence, l = 50;
            quad1: quad, at = 5;
            quad2: quad, at = 10;
            quad3: quad, at = 20;
            bend1: bend, at = 25;
            sext1: sext, at = 30;
            octu1: octu, at = 35;
            sole1: sole, at = 40;
        endsequence;

        beam;

        use, sequence = seq;

        select, pattern = quad, flag = error;
        select, pattern = bend, flag = error;
        select, pattern = sext, flag = error;
        select, pattern = octu, flag = error;
        select, pattern = sole, flag = error;
        efcomp, radius = radius,
            dknr = {dknr0, dknr1, dknr2, dknr3},
            dksr = {dksr0, dksr1, dksr2, dksr3};
    ''')

    line = xt.Line.from_madx_sequence(
        sequence=mad.sequence.seq,
        enable_field_errors=True,
        allow_thick=True,
    )

    def error(name, which):
        field_errors = mad.sequence.seq.expanded_elements[name].field_errors
        dk = getattr(field_errors, which)
        return dk[:6]

    xo.assert_allclose(line['bend1'].knl, error('bend1', 'dkn'), atol=0, rtol=1e-15)
    xo.assert_allclose(line['bend1'].ksl, error('bend1', 'dks'), atol=0, rtol=1e-15)

    xo.assert_allclose(line['quad1'].knl, error('quad1', 'dkn'), atol=0, rtol=1e-15)
    xo.assert_allclose(line['quad2'].knl, error('quad2', 'dkn'), atol=0, rtol=1e-15)
    xo.assert_allclose(line['quad3'].knl, error('quad3', 'dkn'), atol=0, rtol=1e-15)

    xo.assert_allclose(line['quad1'].ksl, error('quad1', 'dks'), atol=0, rtol=1e-15)
    xo.assert_allclose(line['quad2'].ksl, error('quad2', 'dks'), atol=0, rtol=1e-15)
    xo.assert_allclose(line['quad3'].ksl, error('quad3', 'dks'), atol=0, rtol=1e-15)

    xo.assert_allclose(line['sext1'].knl, error('sext1', 'dkn'), atol=0, rtol=1e-15)
    xo.assert_allclose(line['sext1'].ksl, error('sext1', 'dks'), atol=0, rtol=1e-15)

    xo.assert_allclose(line['octu1'].knl, error('octu1', 'dkn'), atol=0, rtol=1e-15)
    xo.assert_allclose(line['octu1'].ksl, error('octu1', 'dks'), atol=0, rtol=1e-15)

    xo.assert_allclose(line['sole1'].knl, error('sole1', 'dkn'), atol=0, rtol=1e-15)
    xo.assert_allclose(line['sole1'].ksl, error('sole1', 'dks'), atol=0, rtol=1e-15)


def test_matrix():
    mad = Madx(stdout=False)

    mad.input("""
    a11=1;
    m1: matrix,
          rm11:=a11,rm12=2,rm21=3,rm22=4,
          kick1=0.1,kick2=0.2,kick3=0.3;

    ss: sequence, l=1; m1: m1, at=0; endsequence;

    beam; use, sequence=ss;
    """)

    line = MadLoader(mad.sequence.ss).make_line()
    line = MadLoader(mad.sequence.ss, enable_expressions=True).make_line()
    line.vars['a11'] = 2.0
    assert line[1].m1[0, 0] == line.vars['a11']._value


def test_srotation():
    mad = Madx(stdout=False)

    mad.input("""
    angle=0.2;
    rot: srotation,angle:=angle;

    ss: sequence, l=1; rot: rot, at=0; endsequence;

    beam; use, sequence=ss;
    """)

    line = MadLoader(mad.sequence.ss).make_line()
    line = MadLoader(mad.sequence.ss, enable_expressions=True).make_line()
    assert isinstance(line[1], xt.SRotation)
    line.vars['angle'] = 2.0
    assert line[1].angle == line.vars['angle']._value * 180 / np.pi


def test_xrotation():
    mad = Madx(stdout=False)

    mad.input("""
    angle=0.2;
    rot: xrotation,angle:=angle;

    ss: sequence, l=1; rot: rot, at=0; endsequence;

    beam; use, sequence=ss;
    """)

    line = MadLoader(mad.sequence.ss).make_line()
    line = MadLoader(mad.sequence.ss, enable_expressions=True).make_line()
    assert isinstance(line[1], xt.XRotation)
    line.vars['angle'] = 2.0
    assert line[1].angle == line.vars['angle']._value * 180 / np.pi


def test_yrotation():
    mad = Madx(stdout=False)

    mad.input("""
    angle=0.2;
    rot: yrotation,angle:=angle;

    ss: sequence, l=1; rot: rot, at=0; endsequence;

    beam; use, sequence=ss;
    """)

    line = MadLoader(mad.sequence.ss).make_line()
    line = MadLoader(mad.sequence.ss, enable_expressions=True).make_line()
    assert isinstance(line[1], xt.YRotation)
    line.vars['angle'] = 2.0
    assert line[1].angle == line.vars['angle']._value * 180 / np.pi


def test_mad_elements_import():
    mad = Madx(stdout=False)

    # Element definitions
    mad.input("""

    a = 1.;

    cav0: rfcavity, freq:=a*10, lag:=a*0.5, volt:=a*6;
    cav1: rfcavity, lag:=a*0.5, volt:=a*6, harmon:=a*8;
    wire1: wire, current:=a*5, l:=a*0, l_phy:=a*1, l_int:=a*2, xma:=a*1e-3, yma:=a*2e-3;
    mult0: multipole, knl:={a*1,a*2,a*3}, ksl:={0, a*5,a*6}, lrad:=a*1.1;
    mult1: multipole, knl={1,2,3,0}, ksl:={0,2,3};
    mult2: multipole, knl={1,2,3,0}, ksl={0,2,0,0,0};
    mult3: multipole, knl={1,2,3,0}, ksl:={0,2,0,b,0};
    kick0: kicker, hkick:=a*5, vkick:=a*6, lrad:=a*2.2;
    kick1: tkicker, hkick:=a*7, vkick:=a*8, lrad:=a*2.3;
    kick2: hkicker, kick:=a*3, lrad:=a*2.4;
    kick3: vkicker, kick:=a*4, lrad:=a*2.5;
    dipedge0: dipedge, h:=a*0.1, e1:=a*3, fint:=a*4, hgap:=a*0.02;
    rfm0: rfmultipole, volt:=a*2, lag:=a*0.5, freq:=a*100.,
                knl:={a*2,a*3}, ksl:={0,a*5},
                pnl:={a*0.3, a*0.4}, psl:={a*0.5, a*0.6};
    crab0: crabcavity, volt:=a*2, lag:=a*0.5, freq:=a*100.;
    crab1: crabcavity, volt:=a*2, lag:=a*0.5, freq:=a*100., tilt:=a*pi/2;
    oct0: marker, apertype=octagon, aperture:={a * 3, a * 6, a * pi/6, a * pi/3};
    """)

    matrix_m0 = np.random.randn(6) * 1E-6
    matrix_m1 = np.reshape(np.random.randn(36), (6, 6))
    mad.input(f"mat:matrix,l:=0.003*a,"
              f"kick1:={matrix_m0[0]}*a,kick2:={matrix_m0[1]}*a,"
              f"kick3:={matrix_m0[2]}*a,kick4:={matrix_m0[3]}*a,"
              f"kick5:={matrix_m0[4]}*a,kick6:={matrix_m0[5]}*a,"
              f"rm11:={matrix_m1[0, 0]}*a,rm12:={matrix_m1[0, 1]}*a,"
              f"rm13:={matrix_m1[0, 2]}*a,rm14:={matrix_m1[0, 3]}*a,"
              f"rm15:={matrix_m1[0, 4]}*a,rm16:={matrix_m1[0, 5]}*a,"
              f"rm21:={matrix_m1[1, 0]}*a,rm22:={matrix_m1[1, 1]}*a,"
              f"rm23:={matrix_m1[1, 2]}*a,rm24:={matrix_m1[1, 3]}*a,"
              f"rm25:={matrix_m1[1, 4]}*a,rm26:={matrix_m1[1, 5]}*a,"
              f"rm31:={matrix_m1[2, 0]}*a,rm32:={matrix_m1[2, 1]}*a,"
              f"rm33:={matrix_m1[2, 2]}*a,rm34:={matrix_m1[2, 3]}*a,"
              f"rm35:={matrix_m1[2, 4]}*a,rm36:={matrix_m1[2, 5]}*a,"
              f"rm41:={matrix_m1[3, 0]}*a,rm42:={matrix_m1[3, 1]}*a,"
              f"rm43:={matrix_m1[3, 2]}*a,rm44:={matrix_m1[3, 3]}*a,"
              f"rm45:={matrix_m1[3, 4]}*a,rm46:={matrix_m1[3, 5]}*a,"
              f"rm51:={matrix_m1[4, 0]}*a,rm52:={matrix_m1[4, 1]}*a,"
              f"rm53:={matrix_m1[4, 2]}*a,rm54:={matrix_m1[4, 3]}*a,"
              f"rm55:={matrix_m1[4, 4]}*a,rm56:={matrix_m1[4, 5]}*a,"
              f"rm61:={matrix_m1[5, 0]}*a,rm62:={matrix_m1[5, 1]}*a,"
              f"rm63:={matrix_m1[5, 2]}*a,rm64:={matrix_m1[5, 3]}*a,"
              f"rm65:={matrix_m1[5, 4]}*a,rm66={matrix_m1[5, 5]}*a;")

    # Sequence
    mad.input("""
    testseq: sequence, l=10;
    m0: mult0 at=0.1;
    m1: mult1 at=0.11;
    m2: mult2 at=0.12;
    m3: mult3 at=0.13;
    c0: cav0, at=0.2, apertype=circle, aperture=0.01;
    c1: cav1, at=0.2, apertype=circle, aperture=0.01;
    k0: kick0, at=0.3;
    k1: kick1, at=0.33;
    k2: kick2, at=0.34;
    k3: kick3, at=0.35;
    de0: dipedge0, at=0.38;
    r0: rfm0, at=0.4;
    cb0: crab0, at=0.41;
    cb1: crab1, at=0.42;
    w: wire1, at=1;
    mat0:mat, at=2+0.003/2;
    oct: oct0, at=3;
    endsequence;
    """
              )

    # Beam
    mad.input(f"""
    beam, particle=ion, gamma=1.05, mass={xt.PROTON_MASS_EV / 1e9}, sequence=testseq;
    """)

    mad.use('testseq')

    seq = mad.sequence['testseq']

    for test_expressions in [True, False]:
        line = xt.Line.from_madx_sequence(sequence=seq,
                                          deferred_expressions=test_expressions,
                                          install_apertures=True)
        line.particle_ref = xp.Particles(mass0=xp.PROTON_MASS_EV, gamma0=1.05)

        line = xt.Line.from_dict(line.to_dict())  # This calls the to_dict method fot all
        # elements

        assert len(line.element_names) == len(line.element_dict.keys())
        assert line.get_length() == 10

        assert isinstance(line['m0'], xt.Multipole)
        assert line.get_s_position('m0') == 0.1
        assert np.all(line['m0'].knl == np.array([1, 2, 3]))
        assert np.all(line['m0'].ksl == np.array([0, 5, 6]))
        assert line['m0'].hxl == 1
        assert line['m0'].length == 1.1
        assert len(line['m1'].knl) == 3
        assert len(line['m1'].ksl) == 3
        assert len(line['m2'].knl) == 3
        assert len(line['m2'].ksl) == 3
        if test_expressions:
            assert len(line['m3'].knl) == 4
            assert len(line['m3'].ksl) == 4
        else:
            assert len(line['m3'].knl) == 3
            assert len(line['m3'].ksl) == 3

        assert isinstance(line['k0'], xt.Multipole)
        assert line.get_s_position('k0') == 0.3
        assert np.all(line['k0'].knl == np.array([-5]))
        assert np.all(line['k0'].ksl == np.array([6]))
        assert line['k0'].hxl == 0
        assert line['k0'].length == 2.2

        assert isinstance(line['k1'], xt.Multipole)
        assert line.get_s_position('k1') == 0.33
        assert np.all(line['k1'].knl == np.array([-7]))
        assert np.all(line['k1'].ksl == np.array([8]))
        assert line['k1'].hxl == 0
        assert line['k1'].length == 2.3

        assert isinstance(line['k2'], xt.Multipole)
        assert line.get_s_position('k2') == 0.34
        assert np.all(line['k2'].knl == np.array([-3]))
        assert np.all(line['k2'].ksl == np.array([0]))
        assert line['k2'].hxl == 0
        assert line['k2'].length == 2.4

        assert isinstance(line['k3'], xt.Multipole)
        assert line.get_s_position('k3') == 0.35
        assert np.all(line['k3'].knl == np.array([0]))
        assert np.all(line['k3'].ksl == np.array([4]))
        assert line['k3'].hxl == 0
        assert line['k3'].length == 2.5

        assert isinstance(line['c0'], xt.Cavity)
        assert line.get_s_position('c0') == 0.2
        assert line['c0'].frequency == 10e6
        assert line['c0'].lag == 180
        assert line['c0'].voltage == 6e6

        assert isinstance(line['c1'], xt.Cavity)
        assert line.get_s_position('c1') == 0.2
        xo.assert_allclose(line['c1'].frequency, clight * line.particle_ref.beta0 / 10. * 8,
                           rtol=0, atol=1e-7)
        assert line['c1'].lag == 180
        assert line['c1'].voltage == 6e6

        assert isinstance(line['de0'], xt.DipoleEdge)
        assert line.get_s_position('de0') == 0.38
        assert line['de0'].k == 0.1
        assert line['de0'].e1 == 3
        assert line['de0'].fint == 4
        assert line['de0'].hgap == 0.02

        assert isinstance(line['r0'], xt.RFMultipole)
        assert line.get_s_position('r0') == 0.4
        assert np.all(line['r0'].knl == np.array([2, 3]))
        assert np.all(line['r0'].ksl == np.array([0, 5]))
        assert np.all(line['r0'].pn == np.array([0.3 * 360, 0.4 * 360]))
        assert np.all(line['r0'].ps == np.array([0.5 * 360, 0.6 * 360]))
        assert line['r0'].voltage == 2e6
        assert line['r0'].order == 1
        assert line['r0'].frequency == 100e6
        assert line['r0'].lag == 180

        assert isinstance(line['cb0'], xt.RFMultipole)
        assert line.get_s_position('cb0') == 0.41
        assert len(line['cb0'].knl) == 1
        assert len(line['cb0'].ksl) == 1
        xo.assert_allclose(line['cb0'].knl[0], 2 * 1e6 / line.particle_ref.p0c[0],
                           rtol=0, atol=1e-12)
        assert np.all(line['cb0'].ksl == 0)
        assert np.all(line['cb0'].pn == np.array([270]))
        assert np.all(line['cb0'].ps == 0.)
        assert line['cb0'].voltage == 0
        assert line['cb0'].order == 0
        assert line['cb0'].frequency == 100e6
        assert line['cb0'].lag == 0

        assert isinstance(line['cb1'], xt.RFMultipole)
        assert line.get_s_position('cb1') == 0.42
        assert len(line['cb1'].knl) == 1
        assert len(line['cb1'].ksl) == 1
        xo.assert_allclose(line['cb1'].ksl[0], -2 * 1e6 / line.particle_ref.p0c[0],
                           rtol=0, atol=1e-12)
        assert np.all(line['cb1'].knl == 0)
        assert np.all(line['cb1'].ps == np.array([270]))
        assert np.all(line['cb1'].pn == 0.)
        assert line['cb1'].voltage == 0
        assert line['cb1'].order == 0
        assert line['cb1'].frequency == 100e6
        assert line['cb1'].lag == 0

        assert isinstance(line['w'], xt.Wire)
        assert line.get_s_position('w') == 1
        assert line['w'].L_phy == 1
        assert line['w'].L_int == 2
        assert line['w'].xma == 1e-3
        assert line['w'].yma == 2e-3

        assert isinstance(line['mat0'], xt.FirstOrderTaylorMap)
        assert line.get_s_position('mat0') == 2
        xo.assert_allclose(line['mat0'].m0, matrix_m0, rtol=0.0, atol=1E-12)
        xo.assert_allclose(line['mat0'].m1, matrix_m1, rtol=0.0, atol=1E-12)

        assert isinstance(line['oct_aper'], xt.LimitPolygon)
        assert line.get_s_position('oct_aper') == 3
        x_1, x_2, y_1, y_2 = 3, 2 * np.sqrt(3), np.sqrt(3), 6
        expected_x_vertices = [x_1, x_2, -x_2, -x_1, -x_1, -x_2, x_2, x_1]
        expected_y_vertices = [y_1, y_2, y_2, y_1, -y_1, -y_2, -y_2, -y_1]
        xo.assert_allclose(line['oct_aper'].x_vertices, expected_x_vertices)
        xo.assert_allclose(line['oct_aper'].y_vertices, expected_y_vertices)


def test_selective_expr_import_and_replace_in_expr():
    # Load line with knobs on correctors only
    from cpymad.madx import Madx
    mad = Madx(stdout=False)
    mad.call(str(test_data_folder /
                 'hllhc14_no_errors_with_coupling_knobs/lhcb1_seq.madx'))
    mad.use(sequence='lhcb1')
    line = xt.Line.from_madx_sequence(mad.sequence.lhcb1,
                                      deferred_expressions=True,
                                      replace_in_expr={'bv_aux': 'bv_aux_lhcb1'},
                                      expressions_for_element_types=('kicker', 'hkicker', 'vkicker'))

    assert len(line.vars['bv_aux_lhcb1']._find_dependant_targets()) > 1
    assert 'bv_aux' not in line.vars

    assert line.element_refs['mqxfa.b3r5..1'].knl[1]._expr is None  # multipole
    assert line.element_refs['mcbxfbv.b2r1'].ksl[0]._expr is not None  # kicker


def test_load_madx_optics_file():
    collider = xt.Environment.from_json(
        test_data_folder / 'hllhc15_thick/hllhc15_collider_thick.json')
    collider.build_trackers()

    # Check varval behaviour
    collider.vars['on_x1'] = 40
    assert collider.varval['on_x1'] == 40
    assert collider.lhcb1.vars['on_x1']._value == 40
    assert collider.lhcb2.vars['on_x1']._value == 40
    assert collider.lhcb1.varval['on_x1'] == 40
    assert collider.lhcb2.varval['on_x1'] == 40

    collider.varval['on_x1'] = 50
    assert collider.vars['on_x1']._value == 50
    assert collider.lhcb1.vars['on_x1']._value == 50
    assert collider.lhcb2.vars['on_x1']._value == 50
    assert collider.lhcb1.varval['on_x1'] == 50
    assert collider.lhcb2.varval['on_x1'] == 50

    collider.lhcb1.varval['on_x1'] = 60
    assert collider.vars['on_x1']._value == 60
    assert collider.lhcb1.vars['on_x1']._value == 60
    assert collider.lhcb2.vars['on_x1']._value == 60
    assert collider.lhcb1.varval['on_x1'] == 60
    assert collider.lhcb2.varval['on_x1'] == 60

    collider.lhcb2.varval['on_x1'] = 70
    assert collider.vars['on_x1']._value == 70
    assert collider.lhcb1.vars['on_x1']._value == 70
    assert collider.lhcb2.vars['on_x1']._value == 70
    assert collider.lhcb1.varval['on_x1'] == 70
    assert collider.lhcb2.varval['on_x1'] == 70

    collider.vars['on_disp'] = 0  # more precise angle
    xo.assert_allclose(collider.twiss().lhcb1['px', 'ip1'], 70e-6, atol=1e-8, rtol=0)
    xo.assert_allclose(collider.twiss().lhcb2['px', 'ip1'], -70e-6, atol=1e-8, rtol=0)

    collider.vars.load_madx_optics_file(
        test_data_folder / 'hllhc15_thick/opt_round_300_1500.madx')

    collider._xdeps_manager.verify()

    tw = collider.twiss()
    xo.assert_allclose(tw.lhcb1.qx, 62.31000000, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb1.qy, 60.32000000, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb2.qx, 62.31000000, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb2.qy, 60.32000000, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb1['betx', 'ip1'], 0.30, atol=0, rtol=1e-4)
    xo.assert_allclose(tw.lhcb1['bety', 'ip1'], 0.30, atol=0, rtol=1e-4)
    xo.assert_allclose(tw.lhcb2['betx', 'ip1'], 0.30, atol=0, rtol=1e-4)
    xo.assert_allclose(tw.lhcb2['bety', 'ip1'], 0.30, atol=0, rtol=1e-4)

    # Check a knob
    collider.vars['on_x1'] = 30
    collider.vars['on_disp'] = 0
    tw = collider.twiss()
    xo.assert_allclose(tw.lhcb1['px', 'ip1'], 30e-6, atol=1e-9, rtol=0)
    xo.assert_allclose(tw.lhcb2['px', 'ip1'], -30e-6, atol=1e-9, rtol=0)

    collider.vars.load_madx_optics_file(
        test_data_folder / 'hllhc15_thick/opt_round_150_1500.madx')

    tw = collider.twiss()
    xo.assert_allclose(tw.lhcb1.qx, 62.31000000, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb1.qy, 60.32000000, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb2.qx, 62.31000000, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb2.qy, 60.32000000, atol=1e-6, rtol=0)
    xo.assert_allclose(tw.lhcb1['betx', 'ip1'], 0.15, atol=0, rtol=1e-6)
    xo.assert_allclose(tw.lhcb1['bety', 'ip1'], 0.15, atol=0, rtol=1e-6)
    xo.assert_allclose(tw.lhcb2['betx', 'ip1'], 0.15, atol=0, rtol=1e-6)
    xo.assert_allclose(tw.lhcb2['bety', 'ip1'], 0.15, atol=0, rtol=1e-6)

    # Check a knob
    collider.vars['on_x1'] = 10
    collider.vars['on_disp'] = 0
    tw = collider.twiss()
    xo.assert_allclose(tw.lhcb1['px', 'ip1'], 10e-6, atol=1e-9, rtol=0)
    xo.assert_allclose(tw.lhcb2['px', 'ip1'], -10e-6, atol=1e-9, rtol=0)

    # Try unregister/register

    collider.vars['ox_x1h'] = 20
    tw = collider.twiss()
    xo.assert_allclose(tw.lhcb1['px', 'ip1'], 10e-6, atol=1e-9, rtol=0)
    xo.assert_allclose(tw.lhcb2['px', 'ip1'], -10e-6, atol=1e-9, rtol=0)

    collider.vars['on_x8h'] = collider.vars['on_x2']
    collider.vars['on_x2'] = 25
    tw = collider.twiss()

    xo.assert_allclose(tw.lhcb1['px', 'ip8'], 25e-6, atol=1e-9, rtol=0)
    xo.assert_allclose(tw.lhcb2['px', 'ip8'], -25e-6, atol=1e-9, rtol=0)
    xo.assert_allclose(tw.lhcb1['py', 'ip8'], 0, atol=1e-9, rtol=0)
    xo.assert_allclose(tw.lhcb2['py', 'ip8'], 0, atol=1e-9, rtol=0)
    xo.assert_allclose(tw.lhcb1['py', 'ip2'], 25e-6, atol=1e-9, rtol=0)
    xo.assert_allclose(tw.lhcb2['py', 'ip2'], -25e-6, atol=1e-9, rtol=0)
    xo.assert_allclose(tw.lhcb1['px', 'ip2'], 0, atol=1e-9, rtol=0)
    xo.assert_allclose(tw.lhcb2['px', 'ip2'], 0, atol=1e-9, rtol=0)


def test_load_b2_with_bv_minus_one():
    test_data_folder_str = str(test_data_folder)

    mad1 = Madx(stdout=False)
    mad1.call(test_data_folder_str + '/hllhc15_thick/lhc.seq')
    mad1.call(test_data_folder_str + '/hllhc15_thick/hllhc_sequence.madx')
    mad1.input('beam, sequence=lhcb1, particle=proton, energy=7000;')
    mad1.use('lhcb1')
    mad1.input('beam, sequence=lhcb2, particle=proton, energy=7000, bv=-1;')
    mad1.use('lhcb2')
    mad1.call(test_data_folder_str + '/hllhc15_thick/opt_round_150_1500.madx')
    mad1.twiss()

    mad4 = Madx(stdout=False)
    mad4.input('mylhcbeam=4')
    mad4.call(test_data_folder_str + '/hllhc15_thick/lhcb4.seq')
    mad4.call(test_data_folder_str + '/hllhc15_thick/hllhc_sequence.madx')
    mad4.input('beam, sequence=lhcb2, particle=proton, energy=7000;')
    mad4.use('lhcb2')
    mad4.call(test_data_folder_str + '/hllhc15_thick/opt_round_150_1500.madx')
    mad4.twiss()

    for mad in [mad1, mad4]:
        mad.globals['vrf400'] = 16  # Check voltage expressions
        mad.globals['lagrf400.b2'] = 0.02  # Check lag expressions
        mad.globals['on_x1'] = 100  # Check kicker expressions
        mad.globals['on_sep2'] = 2  # Check kicker expressions
        mad.globals['on_x5'] = 123  # Check kicker expressions
        mad.globals['kqtf.b2'] = 1e-5  # Check quad expressions
        mad.globals['ksf.b2'] = 1e-3  # Check sext expressions
        mad.globals['kqs.l3b2'] = 1e-4  # Check skew expressions
        mad.globals['kss.a45b2'] = 1e-4  # Check skew sext expressions
        mad.globals['kof.a34b2'] = 3  # Check oct expressions
        mad.globals['on_crab1'] = -190  # Check cavity expressions
        mad.globals['on_crab5'] = -130  # Check cavity expressions
        mad.globals['on_sol_atlas'] = 1  # Check solenoid expressions
        mad.globals['kcdx3.r1'] = 1e-4  # Check thin decapole expressions
        mad.globals['kcdsx3.r1'] = 1e-4  # Check thin skew decapole expressions
        mad.globals['kctx3.l1'] = 1e-5  # Check thin dodecapole expressions
        mad.globals['kctsx3.r1'] = 1e-5  # Check thin skew dodecapole expressions

    line2 = xt.Line.from_madx_sequence(mad1.sequence.lhcb2,
                                       allow_thick=True,
                                       deferred_expressions=True,
                                       replace_in_expr={'bv_aux': 'bvaux_b2'})
    line2.particle_ref = xp.Particles(mass0=xp.PROTON_MASS_EV, p0c=7000e9)

    line4 = xt.Line.from_madx_sequence(mad4.sequence.lhcb2,
                                       allow_thick=True,
                                       deferred_expressions=True,
                                       replace_in_expr={'bv_aux': 'bvaux_b2'})
    line4.particle_ref = xp.Particles(mass0=xp.PROTON_MASS_EV, p0c=7000e9)

    # Bend done

    # Quadrupole
    xo.assert_allclose(line2['mq.27l2.b2'].k1, line4['mq.27l2.b2'].k1, rtol=0, atol=1e-12)
    xo.assert_allclose(line2['mqs.27l3.b2'].k1s, line4['mqs.27l3.b2'].k1s, rtol=0, atol=1e-12)

    tt2 = line2.get_table()
    tt4 = line4.get_table()

    tt2nodr = tt2.rows[tt2.element_type != 'Drift']
    tt4nodr = tt4.rows[tt4.element_type != 'Drift']

    # Check s
    l2names = list(tt2nodr.name)
    l4names = list(tt4nodr.name)

    l2names.remove('lhcb2$start')
    l2names.remove('lhcb2$end')
    l4names.remove('lhcb2$start')
    l4names.remove('lhcb2$end')

    assert set(l2names) == set(l4names)

    xo.assert_allclose(
        tt2nodr.rows[l2names].s, tt4nodr.rows[l2names].s, rtol=0, atol=1e-8)

    for nn in l2names:
        print(nn + '              ', end='\r', flush=True)
        if nn == '_end_point':
            continue
        e2 = line2[nn]
        e4 = line4[nn]
        d2 = e2.to_dict()
        d4 = e4.to_dict()
        for kk in d2.keys():
            if kk in ('__class__', 'model', 'side'):
                assert d2[kk] == d4[kk]
                continue
            xo.assert_allclose(d2[kk], d4[kk], rtol=1e-10, atol=1e-16)



import pathlib

import numpy as np
import pytest

import xobjects as xo
import xtrack as xt
from xobjects.test_helpers import fix_random_seed

test_data_folder = pathlib.Path(
        __file__).parent.joinpath('../test_data').absolute()

configurations = [
    {
        'wiggler_on': False,
        'vertical_orbit_distortion': False,
        'tilt_machine_by_90_degrees': False,
        'check_against_tracking': False,
    },
    {
        'wiggler_on': False,
        'vertical_orbit_distortion': False,
        'tilt_machine_by_90_degrees': True,
        'check_against_tracking': False,
    },
    {
        'wiggler_on': False,
        'vertical_orbit_distortion': True,
        'tilt_machine_by_90_degrees': False,
        'check_against_tracking': True,
    },
    {
        'wiggler_on': False,
        'vertical_orbit_distortion': True,
        'tilt_machine_by_90_degrees': True,
        'check_against_tracking': False,
    },
    {
        'wiggler_on': True,
        'vertical_orbit_distortion': False,
        'tilt_machine_by_90_degrees': False,
        'check_against_tracking': False,
    },
    {
        'wiggler_on': True,
        'vertical_orbit_distortion': False,
        'tilt_machine_by_90_degrees': True,
        'check_against_tracking': True,
    },
]


@pytest.mark.parametrize('conf', configurations)
@fix_random_seed(856384)
def test_eq_emitt(conf):

    test_context = xo.context_default # On GPU this is too slow to run routinely

    print('===============================')
    print(conf)
    print('===============================')

    tilt_machine_by_90_degrees = conf['tilt_machine_by_90_degrees']
    wiggler_on = conf['wiggler_on']
    vertical_orbit_distortion = conf['vertical_orbit_distortion']

    line = xt.Line.from_json(test_data_folder / 'fcc_ee/fccee_h_thick.json')
    line.build_tracker()

    # Wiggler is very strong --> needs more twiss points for accurate eq emittances
    line.slice_thick_elements(slicing_strategies=[
        xt.Strategy(slicing=None), # Default
        xt.Strategy(slicing=xt.Teapot(20, mode='thick'), name=r'^mwi.*'),
    ])
    line.build_tracker()

    print('Done building tracker')

    if wiggler_on:
        line.vars['on_wiggler_v'] = 0.4

    if vertical_orbit_distortion:
        line['mwi.e5rg'].knl[0] = 2e-7

    if tilt_machine_by_90_degrees:

        tw_before_tilt = line.twiss()

        # Bring the machine to the vertical plane
        already_tilted = []
        for nn in line.element_names:
            if hasattr(line[nn], 'rot_s_rad'):
                if nn not in already_tilted:
                    line[nn].rot_s_rad += np.pi/2
                    already_tilted.append(nn)
            if hasattr(line[nn], 'parent_name'):
                nn_parent = line[nn].parent_name
                if hasattr(line[nn_parent], 'rot_s_rad'):
                    if nn_parent not in already_tilted:
                        line[nn_parent].rot_s_rad += np.pi/2
                        already_tilted.append(nn_parent)

        tw_after_tilt = line.twiss()

        assert_allclose = np.testing.assert_allclose

        assert_allclose(tw_after_tilt.qy, tw_before_tilt.qx, rtol=0, atol=1e-8)
        assert_allclose(tw_after_tilt.qx, tw_before_tilt.qy, rtol=0, atol=1e-8)
        assert_allclose(tw_after_tilt.dqy, tw_before_tilt.dqx, rtol=0, atol=1e-4)
        assert_allclose(tw_after_tilt.dqx, tw_before_tilt.dqy, rtol=0, atol=1e-4)

        assert_allclose(tw_after_tilt.bety, tw_before_tilt.betx, rtol=3e-5, atol=0)
        assert_allclose(tw_after_tilt.betx, tw_before_tilt.bety, rtol=3e-5, atol=0)

        assert_allclose(tw_after_tilt.y, tw_before_tilt.x, rtol=0, atol=1e-9)
        assert_allclose(tw_after_tilt.x, -tw_before_tilt.y, rtol=0, atol=1e-9)

        assert_allclose(tw_after_tilt.dy, tw_before_tilt.dx, rtol=0, atol=5e-6)
        assert_allclose(tw_after_tilt.dx, -tw_before_tilt.dy, rtol=0, atol=5e-6)

    line.configure_radiation(model='mean')
    line.compensate_radiation_energy_loss()

    tw_rad = line.twiss(eneloss_and_damping=True)
    ex = tw_rad.eq_gemitt_x
    ey = tw_rad.eq_gemitt_y
    ez = tw_rad.eq_gemitt_zeta

    # for regression testing
    checked = False
    if not tilt_machine_by_90_degrees and not vertical_orbit_distortion and not wiggler_on:
        xo.assert_allclose(ex, 7.1357e-10, atol=0,     rtol=1e-4)
        xo.assert_allclose(ey, 0,          atol=1e-14, rtol=0)
        xo.assert_allclose(ez, 3.6000e-6,  atol=0,     rtol=1e-4)
        checked = True
    elif tilt_machine_by_90_degrees and not vertical_orbit_distortion and not wiggler_on:
        xo.assert_allclose(ex, 0,          atol=1e-14, rtol=0)
        xo.assert_allclose(ey, 7.1358e-10, atol=0,     rtol=1e-4)
        xo.assert_allclose(ez, 3.6000e-6,  atol=0,     rtol=1e-4)
        checked = True
    elif not tilt_machine_by_90_degrees and not vertical_orbit_distortion and wiggler_on:
        xo.assert_allclose(ex, 7.0714e-10, atol=0,     rtol=1e-4)
        xo.assert_allclose(ey, 5.6804e-13, atol=0,     rtol=4e-3)
        xo.assert_allclose(ez, 3.8595e-6,  atol=0,     rtol=1e-4)
        checked = True
    elif tilt_machine_by_90_degrees and not vertical_orbit_distortion and wiggler_on:
        xo.assert_allclose(ex, 5.7068e-13, atol=0,     rtol=1e-2)  # Quite large, to be kept in mind
        xo.assert_allclose(ey, 7.0714e-10, atol=0,     rtol=1e-4)
        xo.assert_allclose(ez, 3.8595e-6,  atol=0,     rtol=1e-4)
        checked = True
    elif not tilt_machine_by_90_degrees and vertical_orbit_distortion and not wiggler_on:
        xo.assert_allclose(ex, 7.1345e-10, atol=0,     rtol=1e-4)
        xo.assert_allclose(ey, 2.2295e-12, atol=0,     rtol=7e-3)
        xo.assert_allclose(ez, 3.5828e-6,  atol=0,     rtol=1e-4)
        checked = True
    elif tilt_machine_by_90_degrees and vertical_orbit_distortion and not wiggler_on:
        xo.assert_allclose(ex, 2.2261e-12, atol=0,     rtol=5e-3)
        xo.assert_allclose(ey, 7.1345e-10, atol=0,     rtol=1e-4)
        xo.assert_allclose(ez, 3.5828e-6,  atol=0,     rtol=1e-4)
        checked = True
    else:
        raise ValueError('Unknown configuration')

    assert checked

    # Check radiation integrals
    tw_integ = line.twiss(radiation_integrals=True)
    xo.assert_allclose(tw_integ.rad_int_damping_constant_x_s,
                       tw_rad.damping_constants_s[0], rtol=0.02, atol=0)
    xo.assert_allclose(tw_integ.rad_int_damping_constant_y_s,
                       tw_rad.damping_constants_s[1], rtol=0.02, atol=0)
    xo.assert_allclose(tw_integ.rad_int_damping_constant_zeta_s,
                       tw_rad.damping_constants_s[2], rtol=0.02, atol=0)
    xo.assert_allclose(tw_integ.rad_int_eq_gemitt_x, ex, rtol=0.15, atol=1e-14)
    xo.assert_allclose(tw_integ.rad_int_eq_gemitt_y, ey, rtol=0.15, atol=1e-14)

    tw_rad2 = line.twiss(eneloss_and_damping=True, method='6d',
                     radiation_method='full',
                     compute_lattice_functions=False,
                     compute_chromatic_properties=False)

    assert 'x' in tw_rad2
    assert 'betx' not in tw_rad2
    assert 'circumference' in tw_rad2
    assert 'qx' not in tw_rad2
    assert 'dqx' not in tw_rad2

    if not vertical_orbit_distortion: # Known inconsistency to be investigated
        xo.assert_allclose(tw_rad2.eq_gemitt_x, tw_rad.eq_gemitt_x, atol=1e-14, rtol=1.5e-2)
        xo.assert_allclose(tw_rad2.eq_gemitt_y, tw_rad.eq_gemitt_y, atol=1e-14, rtol=1.5e-2)
        xo.assert_allclose(tw_rad2.eq_gemitt_zeta, tw_rad.eq_gemitt_zeta, atol=1e-14, rtol=4e-2)
        xo.assert_allclose(tw_rad2.eq_nemitt_x/tw_rad.gamma0, tw_rad.eq_nemitt_x/tw_rad.gamma0, atol=1e-14, rtol=1.5e-2)
        xo.assert_allclose(tw_rad2.eq_nemitt_y/tw_rad.gamma0, tw_rad.eq_nemitt_y/tw_rad.gamma0, atol=1e-14, rtol=1.5e-2)
        xo.assert_allclose(tw_rad2.eq_nemitt_zeta/tw_rad.gamma0, tw_rad.eq_nemitt_zeta/tw_rad.gamma0, atol=1e-14, rtol=4e-2)

    xo.assert_allclose(tw_rad.eq_nemitt_x, tw_rad.eq_gemitt_x * (tw_rad.gamma0*tw_rad.beta0), atol=1e-16, rtol=0)
    xo.assert_allclose(tw_rad.eq_nemitt_y, tw_rad.eq_gemitt_y * (tw_rad.gamma0*tw_rad.beta0), atol=1e-16, rtol=0)
    xo.assert_allclose(tw_rad.eq_nemitt_zeta, tw_rad.eq_gemitt_zeta * (tw_rad.gamma0*tw_rad.beta0), atol=1e-16, rtol=0)
    xo.assert_allclose(tw_rad2.eq_nemitt_x, tw_rad2.eq_gemitt_x * (tw_rad2.gamma0*tw_rad2.beta0), atol=1e-16, rtol=0)
    xo.assert_allclose(tw_rad2.eq_nemitt_y, tw_rad2.eq_gemitt_y * (tw_rad2.gamma0*tw_rad2.beta0), atol=1e-16, rtol=0)
    xo.assert_allclose(tw_rad2.eq_nemitt_zeta, tw_rad2.eq_gemitt_zeta * (tw_rad2.gamma0*tw_rad2.beta0), atol=1e-16, rtol=0)

    if conf['check_against_tracking']:

        line.discard_tracker()
        line.build_tracker(_context=test_context)

        line.configure_radiation(model='quantum')
        p = line.build_particles(num_particles=30)
        line.track(p, num_turns=400, turn_by_turn_monitor=True, time=True)
        mon = line.record_last_track
        print(f'Tracking time: {line.time_last_track}')

        sigma_x_eq = float(np.sqrt(ex * tw_rad.betx[0] + ey * tw_rad.betx2[0] + (np.std(p.delta) * tw_rad.dx[0])**2))
        sigma_y_eq = float(np.sqrt(ex * tw_rad.bety1[0] + ey * tw_rad.bety[0] + (np.std(p.delta) * tw_rad.dy[0])**2))
        sigma_zeta_eq = float(np.sqrt(ez * tw_rad.bets0))

        sigma_x_track = np.std(mon.x, axis=0)[-200:]
        sigma_y_track = np.std(mon.y, axis=0)[-200:]
        sigma_zeta_track = np.std(mon.zeta, axis=0)[-200:]

        if sigma_x_eq > 1e-8:
            assert np.min(np.abs(sigma_x_track/sigma_x_eq - 1.)) < 0.1
        if sigma_y_eq > 1e-8:
            assert np.min(np.abs(sigma_y_track/sigma_y_eq - 1.)) < 0.1
        assert np.min(np.abs(sigma_zeta_track/sigma_zeta_eq - 1.)) < 0.1

        xo.assert_allclose(sigma_x_eq, np.mean(sigma_x_track), rtol=0.3, atol=1e-9)
        xo.assert_allclose(sigma_y_eq, np.mean(sigma_y_track), rtol=0.3, atol=1e-9)
        xo.assert_allclose(sigma_zeta_eq, np.mean(sigma_zeta_track), rtol=0.3, atol=1e-9)



import json

import numpy as np
import xtrack as xt
import xobjects as xo

with open('../../test_data/hllhc14_no_errors_with_coupling_knobs/line_b1.json',
            'r') as fid:
    dct = json.load(fid)
line = xt.Line.from_dict(dct)

line.build_tracker()

tw_before = line.twiss()

GreaterThan = xt.GreaterThan
LessThan = xt.LessThan

tw0 = line.twiss()
opt = line.match(
    solve=False,
    solver='jacobian',
    # Portion of the beam line to be modified and initial conditions
    start='mq.33l8.b1',
    end='mq.17l8.b1',
    init=tw0, init_at=xt.START,
    # Dipole corrector strengths to be varied
    vary=[
        xt.Vary(name='acbv32.l8b1', step=1e-10, weight=0.7),
        xt.Vary(name='acbv28.l8b1', step=1e-10, weight=0.3),
        xt.Vary(name='acbv26.l8b1', step=1e-10),
        xt.Vary(name='acbv24.l8b1', step=1e-10),
        xt.Vary(name='acbv22.l8b1', step=1e-10, limits=[-38e-6, 38e-6], weight=1000),
        xt.Vary(name='acbv18.l8b1', step=1e-10),
    ],
    targets=[
        # I want the vertical orbit to be at 3 mm at mq.28l8.b1 with zero angle
        xt.Target('y', at='mb.b26l8.b1', value=3e-3, tol=1e-4),
        xt.Target('py', at='mb.b26l8.b1', value=0, tol=1e-6, weight=1e3),
        # I want the bump to be closed
        xt.Target('y', at='mq.17l8.b1', value=tw0, tol=1e-6),
        xt.Target('py', at='mq.17l8.b1', value=tw0, tol=1e-7, weight=1e3),
    ]
)

opt.solve()

# I want to limit the negative excursion ot the bump
opt2 = opt.clone(
    add_targets=[
        xt.Target('y', GreaterThan(-1e-3), at='mq.30l8.b1', tol=1e-6)])
opt2.solve()

#!end-doc-part

tw = line.twiss()

import matplotlib.pyplot as plt
plt.close('all')
fig = plt.figure(1, figsize=(6.4*1.2, 4.8*0.8))
ax = fig.add_subplot(111)
ax.plot(tw_before.s, tw_before.y*1000, label='y')
ax.plot(tw.s, tw.y*1000, label='y')
# Target
ax.axvline(x=line.get_s_position('mb.b26l8.b1'), color='r', linestyle='--', alpha=0.5)
# Correctors
ax.axvline(x=line.get_s_position('mcbv.32l8.b1'), color='k', linestyle='--', alpha=0.5)
ax.axvline(x=line.get_s_position('mcbv.28l8.b1'), color='k', linestyle='--', alpha=0.5)
ax.axvline(x=line.get_s_position('mcbv.26l8.b1'), color='k', linestyle='--', alpha=0.5)
ax.axvline(x=line.get_s_position('mcbv.24l8.b1'), color='k', linestyle='--', alpha=0.5)
ax.axvline(x=line.get_s_position('mcbv.22l8.b1'), color='k', linestyle='--', alpha=0.5)
ax.axvline(x=line.get_s_position('mcbv.18l8.b1'), color='k', linestyle='--', alpha=0.5)
# Boundaries
ax.axvline(x=line.get_s_position('mq.33l8.b1'), color='g', linestyle='--', alpha=0.5)
ax.axvline(x=line.get_s_position('mq.17l8.b1'), color='g', linestyle='--', alpha=0.5)
ax.axhline(y=-1, color='b', linestyle='--', alpha=0.5)
ax.set_xlim(line.get_s_position('mq.33l8.b1') - 10,
            line.get_s_position('mq.17l8.b1') + 10)
ax.set_xlabel('s [m]')
ax.set_ylabel('y [mm]')
ax.set_ylim(-10, 10)
plt.subplots_adjust(bottom=.152, top=.9, left=.1, right=.95)
plt.show()

assert np.isclose(tw['y', 'mq.33l8.b1'], 0, atol=1e-6, rtol=0)
assert np.isclose(tw['y', 'mq.17l8.b1'], 0, atol=1e-6, rtol=0)
assert np.isclose(tw['py', 'mq.17l8.b1'], 0, atol=1e-8, rtol=0)
assert np.isclose(tw['py', 'mq.33l8.b1'], 0, atol=1e-6, rtol=0)

assert np.isclose(tw['y', 'mb.b26l8.b1'], 3e-3, atol=1e-6, rtol=0)
assert np.isclose(tw['py', 'mb.b26l8.b1'], 0, atol=1e-8, rtol=0)

assert np.isclose(tw['y', 'mq.30l8.b1'], -1e-3, atol=1e-6, rtol=0)
assert np.isclose(line.vars['acbv22.l8b1']._value, 38e-6, atol=0, rtol=0.02)



from cpymad.madx import Madx
import xtrack as xt

import numpy as np


collider = xt.Environment.from_json(
    '../../test_data/hllhc15_thick/hllhc15_collider_thick.json')
collider['lhcb1'].twiss_default['method'] = '4d'
collider['lhcb2'].twiss_default['method'] = '4d'

# Generate MAD-X maps

mad = Madx()
mad.input(f"""
call,file="../../test_data/hllhc15_thick/lhc.seq";
call,file="../../test_data/hllhc15_thick/hllhc_sequence.madx";
beam, sequence=lhcb1, particle=proton, pc=7000;
beam, sequence=lhcb2, particle=proton, pc=7000, bv=-1;
call,file="../../test_data/hllhc15_thick/opt_round_150_1500.madx";
""")

for line_name in ['lhcb1', 'lhcb2']:

    if line_name == 'lhcb1':
        line = collider.lhcb1
        mad.use(sequence="lhcb1")
        seq = mad.sequence.lhcb1
    elif line_name == 'lhcb2':
        line = collider.lhcb2
        mad.use(sequence="lhcb2")
        seq = mad.sequence.lhcb2
        line.twiss_default['reverse'] = True

    tw = line.twiss()
    twmad = mad.twiss(chrom=True)

    tw_test = tw.rows['.*_exit']

    wx_ref = np.interp(tw_test['s'], twmad['s'], twmad['wx'])
    wy_ref = np.interp(tw_test['s'], twmad['s'], twmad['wy'])
    phix_ref = np.interp(tw_test['s'], twmad['s'], twmad['phix'])
    phiy_ref = np.interp(tw_test['s'], twmad['s'], twmad['phiy'])

    zx_ref = wx_ref * np.exp(1j * 2 * np.pi * phix_ref)
    zy_ref = wy_ref * np.exp(1j * 2 * np.pi * phiy_ref)
    ax_ref = np.imag(zx_ref)
    ay_ref = np.imag(zy_ref)
    bx_ref = np.real(zx_ref)
    by_ref = np.real(zy_ref)

    assert np.allclose(tw_test.wx_chrom, wx_ref, rtol=0, atol=2e-3 * np.max(wx_ref))
    assert np.allclose(tw_test.wy_chrom, wy_ref, rtol=0, atol=2e-3 * np.max(wy_ref))
    assert np.allclose(tw_test.ax_chrom, ax_ref, rtol=0, atol=2e-3 * np.max(ax_ref))
    assert np.allclose(tw_test.ay_chrom, ay_ref, rtol=0, atol=2e-3 * np.max(ay_ref))
    assert np.allclose(tw_test.bx_chrom, bx_ref, rtol=0, atol=2e-3 * np.max(bx_ref))
    assert np.allclose(tw_test.by_chrom, by_ref, rtol=0, atol=2e-3 * np.max(by_ref))

    init = tw.get_twiss_init('ip3')
    tw_open = line.twiss(start='ip3', end='ip6', init=init,
                         compute_chromatic_properties=True,
                         )

    tw_ref_open = tw.rows['ip3':'ip6']
    assert np.allclose(tw_open.wx_chrom[:-1], tw_ref_open.wx_chrom,
                       rtol=0, atol=2e-3 * np.max(tw_ref_open.wx_chrom))
    assert np.allclose(tw_open.wy_chrom[:-1], tw_ref_open.wy_chrom,
                       rtol=0, atol=2e-3 * np.max(tw_ref_open.wy_chrom))
    assert np.allclose(tw_open.ax_chrom[:-1], tw_ref_open.ax_chrom,
                          rtol=0, atol=2e-3 * np.max(tw_ref_open.ax_chrom))
    assert np.allclose(tw_open.ay_chrom[:-1], tw_ref_open.ay_chrom,
                            rtol=0, atol=2e-3 * np.max(tw_ref_open.ay_chrom))
    assert np.allclose(tw_open.bx_chrom[:-1], tw_ref_open.bx_chrom,
                            rtol=0, atol=2e-3 * np.max(tw_ref_open.bx_chrom))


import xtrack as xt
import xobjects as xo

xo.context_default.kernels.clear()

knl = [0.01]
ksl = [0]

quad = xt.Quadrupole(k1=0, length=1, knl=knl, ksl=ksl)
d1 = xt.Drift(length=0.5)
m = xt.Multipole(length=1, knl=knl, ksl=ksl)
d2 = xt.Drift(length=0.5)

l2 = xt.Line(elements=[d1, m, d2])
l2.build_tracker()


p0 = xt.Particles(p0c=7000e9, x=1e-3, px=0)

p1 = p0.copy()
p2 = p0.copy()

quad.track(p1)
l2.track(p2)


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xobjects as xo

from ..base_element import BeamElement
from ..general import _pkg_root


UNLIMITED = 1e10  # could use np.inf but better safe than sorry


class LimitRect(BeamElement):

    '''
    Beam element modeling a rectangular aperture limit.

    Parameters
    ----------
    min_x : float
        Lower x limit in meters.
    max_x : float
        Upper x limit in meters.
    min_y : float
        Lower y limit in meters.
    max_y : float
        Upper y limit in meters.

    '''


    _xofields = {
        'min_x': xo.Float64,
        'max_x': xo.Float64,
        'min_y': xo.Float64,
        'max_y': xo.Float64,
        }

    has_backtrack = True

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements/apertures_src/limitrect.h')]

    def __init__(self, min_x=-UNLIMITED, max_x=UNLIMITED, min_y=-UNLIMITED, max_y=UNLIMITED, **kwargs):

        super().__init__(min_x=min_x, max_x=max_x, min_y=min_y, max_y=max_y, **kwargs)


class LimitRacetrack(BeamElement):

    '''
    Beam element modeling a racetrack aperture limit.

    Parameters
    ----------
    min_x : float
        Lower x limit in meters.
    max_x : float
        Upper x limit in meters.
    min_y : float
        Lower y limit in meters.
    max_y : float
        Upper y limit in meters.
    a : float
        Horizontal semi-axis in meters of ellipse used for the rounding of the corners.
    b : float
        Vertical semi-axis in meters of ellipse used for the rounding of the corners.

    '''

    _xofields = {
        'min_x': xo.Float64,
        'max_x': xo.Float64,
        'min_y': xo.Float64,
        'max_y': xo.Float64,
        'a': xo.Float64,
        'b': xo.Float64,
        }

    has_backtrack = True

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements/apertures_src/limitracetrack.h')]

    def __init__(self, min_x=-UNLIMITED, max_x=UNLIMITED, min_y=-UNLIMITED,
                 max_y=UNLIMITED, a=0, b=0, **kwargs):

        if "_xobject" in kwargs:
            self.xoinitialize(_xobject=kwargs['_xobject'])
            return

        assert a >= 0
        assert b >= 0
        assert max_x >= min_x
        assert max_y >= min_y

        if a > 0.5 * (max_x - min_x) or b > 0.5 * (max_y - min_y):
            raise ValueError(
                f"Radii of corners ({a} and {b}) are larger than rectangular limit "
                f"([{min_x}, {max_x}] and [{min_y}, {max_y}])!")

        super().__init__(min_x=min_x, max_x=max_x, min_y=min_y, max_y=max_y, a=a, b=b, **kwargs)


class LimitEllipse(BeamElement):

    '''
    Beam element modeling an elliptical aperture limit.

    Parameters
    ----------
    a : float
        Horizontal semi-axis in meters.
    b : float
        Vertical semi-axis in meters.

    '''

    _xofields = {
            'a_squ': xo.Float64,
            'b_squ': xo.Float64,
            'a_b_squ': xo.Float64,
            }

    has_backtrack = True

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements/apertures_src/limitellipse.h')]

    def to_dict(self):
        dct = super().to_dict()
        dct['a'] = np.sqrt(self.a_squ)
        dct['b'] = np.sqrt(self.b_squ)
        return dct

    def __init__(self, a=None, b=None, a_squ=None, b_squ=None, **kwargs):


        if a is None and a_squ is None:
            a = UNLIMITED

        if b is None and b_squ is None:
            b = UNLIMITED

        if a is not None:
            a_squ = a * a

        if b is not None:
            b_squ = b * b

        if 'a_b_squ' not in kwargs.keys():
            kwargs['a_b_squ'] = a_squ * b_squ

        if a_squ > 0.0 and b_squ > 0.0:
            super().__init__(a_squ=a_squ, b_squ=b_squ,
                             **kwargs)
        else:
            raise ValueError("a_squ and b_squ have to be positive definite")

    @property
    def a(self):
        return np.sqrt(self.a_squ)

    @a.setter
    def a(self, a):
        self.a_squ = a * a
        self.a_b_squ = self.a_squ * self.b_squ

    @property
    def b(self):
        return np.sqrt(self.b_squ)

    @b.setter
    def b(self, b):
        self.b_squ = b * b
        self.a_b_squ = self.a_squ * self.b_squ

    def set_half_axes(self, a, b):
        return self.set_half_axes_squ(a * a, b * b)

    def set_half_axes_squ(self, a_squ, b_squ):
        self.a_squ = a_squ
        self.b_squ = b_squ
        self.a_b_squ = a_squ * b_squ
        return self

    @property
    def _repr_fields(self):
        return ['a', 'b']


class LimitPolygon(BeamElement):

    '''
    Beam element modeling a polygonal aperture limit.

    Parameters
    ----------
    x_vertices : array_like
        x coordinates of the vertices of the polygon in meters.
    y_vertices : array_like
        y coordinates of the vertices of the polygon in meters.

    Notes
    -----
    The polygon is closed automatically by connecting the last and first vertex.

    '''

    _xofields = {
        'x_vertices': xo.Float64[:],
        'y_vertices': xo.Float64[:],
        'x_normal': xo.Float64[:],
        'y_normal': xo.Float64[:],
        'resc_fac': xo.Float64
        }

    has_backtrack = True

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements/apertures_src/limitpolygon.h')]

    _kernels = {
        'LimitPolygon_impact_point_and_normal': xo.Kernel(
            c_name='LimitPolygon_impact_point_and_normal',
            args=[xo.Arg(xo.ThisClass, name='el'),
                  xo.Arg(xo.Float64, pointer=True, name='x_in'),
                  xo.Arg(xo.Float64, pointer=True, name='y_in'),
                  xo.Arg(xo.Float64, pointer=True, name='z_in'),
                  xo.Arg(xo.Float64, pointer=True, name='x_out'),
                  xo.Arg(xo.Float64, pointer=True, name='y_out'),
                  xo.Arg(xo.Float64, pointer=True, name='z_out'),
                  xo.Arg(xo.Int64,   pointer=False, name='n_impacts'),
                  xo.Arg(xo.Float64, pointer=True, name='x_inters'),
                  xo.Arg(xo.Float64, pointer=True, name='y_inters'),
                  xo.Arg(xo.Float64, pointer=True, name='z_inters'),
                  xo.Arg(xo.Float64, pointer=True, name='Nx_inters'),
                  xo.Arg(xo.Float64, pointer=True, name='Ny_inters'),
                  xo.Arg(xo.Int64,   pointer=True, name='i_found')],
            n_threads='n_impacts')}

    def __init__(self, x_vertices=None, y_vertices=None, **kwargs):


        if '_xobject' in kwargs.keys():
            super().__init__(**kwargs)
        else:
            assert len(x_vertices) == len(y_vertices)

            if 'x_normal' not in kwargs.keys():
                kwargs['x_normal'] = len(x_vertices)

            if 'y_normal' not in kwargs.keys():
                kwargs['y_normal'] = len(x_vertices)

            if 'resc_fac' not in kwargs.keys():
                kwargs['resc_fac'] = 1.

            super().__init__(
                    x_vertices=x_vertices,
                    y_vertices=y_vertices,
                    **kwargs)

            lengths = np.sqrt(np.diff(self.x_closed)**2
                              + np.diff(self.y_closed)**2)

            assert np.all(lengths>0)


            if self.area < 0:
                self.x_vertices = self.x_vertices[::-1]
                self.y_vertices = self.y_vertices[::-1]
                #raise ValueError(
                #        "The area of the polygon is negative!\n"
                #        "Vertices must be provided with counter-clockwise order!")

            Nx = -np.diff(self.y_closed)
            Ny = np.diff(self.x_closed)

            norm_N = np.sqrt(Nx**2 + Ny**2)
            Nx = Nx / norm_N
            Ny = Ny / norm_N

            ctx = self._buffer.context
            self.x_normal = ctx.nparray_to_context_array(Nx)
            self.y_normal = ctx.nparray_to_context_array(Ny)

    @property
    def x_closed(self):
        ctx = self._buffer.context
        xx = ctx.nparray_from_context_array(self.x_vertices)
        return np.concatenate([xx, np.array([xx[0]])])

    @property
    def y_closed(self):
        ctx = self._buffer.context
        yy = ctx.nparray_from_context_array(self.y_vertices)
        return np.concatenate([yy, np.array([yy[0]])])

    def impact_point_and_normal(self, x_in, y_in, z_in,
                                x_out, y_out, z_out):

        ctx = self._buffer.context

        if 'LimitPolygon_impact_point_and_normal' not in ctx.kernels.keys():
            # The tracking kernel requires the usual particle class
            self.compile_kernels(only_if_needed=True)

        x_inters = ctx.zeros(shape=x_in.shape, dtype=np.float64)
        y_inters = ctx.zeros(shape=x_in.shape, dtype=np.float64)
        z_inters = ctx.zeros(shape=x_in.shape, dtype=np.float64)
        Nx_inters = ctx.zeros(shape=x_in.shape, dtype=np.float64)
        Ny_inters = ctx.zeros(shape=x_in.shape, dtype=np.float64)
        i_found = ctx.zeros(shape=x_in.shape, dtype=np.int64)

        ctx.kernels.LimitPolygon_impact_point_and_normal(el=self,
                x_in=x_in, y_in=y_in, z_in=z_in,
                x_out=x_out, y_out=y_out, z_out=z_out,
                n_impacts=len(x_in), x_inters=x_inters,
                y_inters=y_inters, z_inters=z_inters,
                Nx_inters=Nx_inters, Ny_inters=Ny_inters,
                i_found=i_found)

        assert np.all(i_found>=0)

        return x_inters, y_inters, z_inters, Nx_inters, Ny_inters, i_found

    @property
    def area(self):
        return -0.5 * np.sum((self.y_closed[1:] + self.y_closed[:-1])
                                * (self.x_closed[1:] - self.x_closed[:-1]))
    @property
    def centroid(self):
        x = self.x_vertices
        y = self.x_vertices
        cx = 1/(6*self.area)*np.sum((x[:-1]+x[1:])*(x[:-1]*y[1:]-x[1:]*y[:-1]))
        cy = 1/(6*self.area)*np.sum((y[:-1]+y[1:])*(y[:-1]*x[1:]-y[1:]*x[:-1]))
        return (cx,cy)

class LimitRectEllipse(BeamElement):

    '''
    Element modeling an aperture limit given by the intersection of
    a symmetric LimitRect and a LimitEllipse.

    The particles are lost if they exceed either the rect or ellipse aperture.

    Parameters
    ----------
    max_x : float
        Horizontal semi-axis of rect in meters.
    max_y : float
        Vertical semi-axis of rect in meters.
    a : float
        Horizontal semi-axis of ellipse in meters.
    b : float
        Vertical semi-axis of ellipse in meters.
    '''

    _xofields = {
            'max_x': xo.Float64,
            'max_y': xo.Float64,
            'a_squ': xo.Float64,
            'b_squ': xo.Float64,
            'a_b_squ': xo.Float64,
            }

    has_backtrack = True

    def __init__(
        self, max_x=UNLIMITED, max_y=UNLIMITED, a_squ=None, b_squ=None,
        a=None, b=None, **kwargs
    ):

        if a is None and a_squ is None:
            a = UNLIMITED

        if b is None and b_squ is None:
            b = UNLIMITED

        if a is not None:
            a_squ = a * a

        if b is not None:
            b_squ = b * b

        if max_x < 0.0:
            raise ValueError("max_x has to be positive definite")

        if max_y < 0.0:
            raise ValueError("max_y has to be_positive definite")

        if a_squ < 0.0 or b_squ < 0.0:
            raise ValueError("a_squ and b_squ have to be positive definite")

        if 'a_b_squ' not in kwargs.keys():
            kwargs['a_b_squ'] = a_squ * b_squ

        super().__init__(
            max_x=max_x,
            max_y=max_y,
            a_squ=a_squ,
            b_squ=b_squ,
            **kwargs
        )

    @property
    def a(self):
        return np.sqrt(self.a_squ)

    @a.setter
    def a(self, a):
        self.a_squ = a * a
        self.a_b_squ = self.a_squ * self.b_squ

    @property
    def b(self):
        return np.sqrt(self.b_squ)

    @b.setter
    def b(self, b):
        self.b_squ = b * b
        self.a_b_squ = self.a_squ * self.b_squ

    def set_half_axes(self, a, b):
        return self.set_half_axes_squ(a * a, b * b)

    def set_half_axes_squ(self, a_squ, b_squ):
        self.a_squ = a_squ
        self.b_squ = b_squ
        self.a_b_squ = a_squ * b_squ
        return self

    @property
    def _repr_fields(self):
        return ['max_x', 'max_y', 'a', 'b']

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements/apertures_src/limitrectellipse.h')]


class LongitudinalLimitRect(BeamElement):

    '''Beam element introducing a limit on the longitudinal coordinates.

    Parameters
    ----------
    min_zeta : float
        Lower limit on zeta coordinate in meters.
    max_zeta : float
        Upper limit on zeta coordinate in meters.
    min_pzeta : float
        Lower limit on pzeta coordinate.
    max_pzeta : float
        Upper limit on pzeta coordinate.
    '''

    _xofields = {
        'min_zeta': xo.Float64,
        'max_zeta': xo.Float64,
        'min_pzeta': xo.Float64,
        'max_pzeta': xo.Float64,
        }

    has_backtrack = True

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements/apertures_src/longitudinallimitrect.h')]

    def __init__(self, min_zeta=-UNLIMITED, max_zeta=UNLIMITED, min_pzeta=-UNLIMITED, max_pzeta=UNLIMITED, **kwargs):

        super().__init__(min_zeta=min_zeta, max_zeta=max_zeta, min_pzeta=min_pzeta, max_pzeta=max_pzeta, **kwargs)






# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json
import pathlib
import numpy as np

import xobjects as xo
import xtrack as xt


short_test = False # Short line (5 elements)

test_data_folder = pathlib.Path(
        __file__).parent.joinpath('../../test_data').absolute()

fname_line_particles = test_data_folder.joinpath(
                        './hllhc_14/line_and_particle.json')

####################
# Choose a context #
####################

context = xo.ContextCpu()
#context = xo.ContextCupy()
#context = xo.ContextPyopencl('0.0')

#############
# Load file #
#############

with open(fname_line_particles, 'r') as fid:
    input_data = json.load(fid)

##############
# Get a line #
##############

line = xt.Line.from_dict(input_data['line'])
line.particle_ref = xt.Particles(**input_data['particle'])

#################
# Build Tracker #
#################
print('Build tracker...')
freeze_vars = xt.Particles.part_energy_varnames() + ['zeta']
line.build_tracker(_context=context)

line.freeze_longitudinal()

#########
# Twiss #
#########

tw = line.twiss(method='4d')  # <-- Need to choose 4d mode when longitudinal
                                 #     variables are frozen

##################################
# Match a particles distribution #
##################################

particles = xt.build_particles(_context=context, line=line,
                               mode = '4d',  # <--- 4d
                               x_norm=np.linspace(0, 10, 11),
                               nemitt_x=3e-6, nemitt_y=3e-6)

particles_before_tracking = particles.copy()

#########
# Track #
#########
print('Track a few turns...')
n_turns = 10
line.track(particles, num_turns=n_turns)

print('Track again (no compile)')
line.track(particles, num_turns=n_turns)

for vv in ['ptau', 'delta', 'rpp', 'rvv', 'zeta']:
    vv_before = context.nparray_from_context_array(
                        getattr(particles_before_tracking, vv))
    vv_after= context.nparray_from_context_array(
                        getattr(particles, vv))
    assert np.all(vv_before == vv_after)

print('Check passed')



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import time
import numpy as np
from scipy import constants

protonMass = constants.value("proton mass energy equivalent in MeV") * 1e6
from scipy.stats import linregress
from scipy.signal import hilbert
from matplotlib import pyplot as plt
plt.close('all')

import xobjects as xo
import xtrack as xt
xt.enable_pyheadtail_interface()

from PyHEADTAIL.particles.generators import generate_Gaussian6DTwiss
from PyHEADTAIL.particles.slicing import UniformBinSlicer
from PyHEADTAIL.impedances.wakes import WakeTable, WakeField
from PyHEADTAIL.feedback.transverse_damper import TransverseDamper
from PyHEADTAIL.trackers.transverse_tracking import TransverseSegmentMap
from PyHEADTAIL.trackers.longitudinal_tracking import LinearMap
from PyHEADTAIL.trackers.detuners import ChromaticitySegment, AmplitudeDetuningSegment

context = xo.ContextCupy()

nTurn = 5000  # int(1E4)
bunch_intensity = 1.8e11
n_macroparticles = int(1e4)
energy = 7e3  # [GeV]
gamma = energy * 1e9 / protonMass
betar = np.sqrt(1 - 1 / gamma ** 2)
normemit = 1.8e-6
beta_x = 68.9
beta_y = 70.34
Q_x = 0.31
Q_y = 0.32
chroma = -5.0  # 10.0
sigma_4t = 1.2e-9
sigma_z = sigma_4t / 4.0 * constants.c
momentumCompaction = 3.483575072011584e-04
eta = momentumCompaction - 1.0 / gamma ** 2
voltage = 12.0e6
h = 35640
p0 = constants.m_p * betar * gamma * constants.c
Q_s = np.sqrt(constants.e * voltage * eta * h / (2 * np.pi * betar * constants.c * p0))
circumference = 26658.883199999
averageRadius = circumference / (2 * np.pi)
sigma_delta = Q_s * sigma_z / (averageRadius * eta)
beta_s = sigma_z / sigma_delta
emit_s = 4 * np.pi * sigma_z * sigma_delta * p0 / constants.e  # eVs for PyHEADTAIL

n_slices_wakes = 200
limit_z = 3 * sigma_z
wakefile = "wakes/wakeforhdtl_PyZbase_Allthemachine_7000GeV_B1_2021_TeleIndex1_wake.dat"
slicer_for_wakefields = UniformBinSlicer(n_slices_wakes, z_cuts=(-limit_z, limit_z))
waketable = WakeTable(
    wakefile, ["time", "dipole_x", "dipole_y", "quadrupole_x", "quadrupole_y"]
)
wake_field = WakeField(slicer_for_wakefields, waketable)
wake_field.needs_cpu = True
wake_field.needs_hidden_lost_particles = True

damping_time = 7000  # 33.
damper = TransverseDamper(dampingrate_x=damping_time, dampingrate_y=damping_time)
damper.needs_cpu = True
damper.needs_hidden_lost_particles = True
i_oct = 15.
det_xx = 1.4e5 * i_oct / 550.0  # from PTC with ATS optics, telescopic factor 1.0
det_xy = -1.0e5 * i_oct / 550.0

# expected octupole threshold with damper is 273A according to https://indico.cern.ch/event/902528/contributions/3798807/attachments/2010534/3359300/20200327_RunIII_stability_NMounet.pdf
# expected growth rate with damper but without octupole is ~0.3 [$10^{-4}$/turn] (also according to Nicolas' presentation)

checkTurn = 1000

########### PyHEADTAIL part ##############

particles = generate_Gaussian6DTwiss(
    macroparticlenumber=n_macroparticles,
    intensity=bunch_intensity,
    charge=constants.e,
    mass=constants.m_p,
    circumference=circumference,
    gamma=gamma,
    alpha_x=0.0,
    alpha_y=0.0,
    beta_x=beta_x,
    beta_y=beta_y,
    beta_z=beta_s,
    epsn_x=normemit,
    epsn_y=normemit,
    epsn_z=emit_s,
)

x0 = np.copy(particles.x)
px0 = np.copy(particles.xp)
y0 = np.copy(particles.y)
py0 = np.copy(particles.yp)
zeta0 = np.copy(particles.z)
delta0 = np.copy(particles.dp)

chromatic_detuner = ChromaticitySegment(dQp_x=chroma, dQp_y=0.0)
transverse_detuner = AmplitudeDetuningSegment(
    dapp_x=det_xx * p0,
    dapp_y=det_xx * p0,
    dapp_xy=det_xy * p0,
    dapp_yx=det_xy * p0,
    alpha_x=0.0,
    beta_x=beta_x,
    alpha_y=0.0,
    beta_y=beta_y,
)
arc_transverse = TransverseSegmentMap(
    alpha_x_s0=0.0,
    beta_x_s0=beta_x,
    D_x_s0=0.0,
    alpha_x_s1=0.0,
    beta_x_s1=beta_x,
    D_x_s1=0.0,
    alpha_y_s0=0.0,
    beta_y_s0=beta_y,
    D_y_s0=0.0,
    alpha_y_s1=0.0,
    beta_y_s1=beta_y,
    D_y_s1=0.0,
    dQ_x=Q_x,
    dQ_y=Q_y,
    segment_detuners=[chromatic_detuner, transverse_detuner],
)
arc_longitudinal = LinearMap(
    alpha_array=[momentumCompaction], circumference=circumference, Q_s=Q_s
)

turns = np.arange(nTurn)
x = np.zeros(nTurn, dtype=float)
t_pyht_start = time.time()
for turn in range(nTurn):
    time0 = time.time()
    arc_transverse.track(particles)
    arc_longitudinal.track(particles)
    time1 = time.time()
    wake_field.track(particles)
    time2 = time.time()
    damper.track(particles)
    time3 = time.time()
    x[turn] = np.mean(particles.x)
    if turn % 1000 == 0:
        print(
            f"PyHt - turn {turn}: time for arc {time1-time0}s, for wake {time2-time1}s, for damper {time3-time2}s"
        )
t_pyht_end = time.time()
print(f'PyHT full loop: {t_pyht_end - t_pyht_start}s')

x /= np.sqrt(normemit * beta_x / gamma / betar)
plt.figure(0)
plt.plot(turns, x, label=f"{i_oct}A")
iMin = 1000
iMax = nTurn - 1000
if iMin >= iMax:
    iMin = 0
    iMax = nTurn
ampl = np.abs(hilbert(x))
b, a, r, p, stderr = linregress(turns[iMin:iMax], np.log(ampl[iMin:iMax]))
gr_pyht = b
plt.plot(turns, np.exp(a + b * turns), "--k", label=f"{1/b:.3E} turns")
print(f"Growth rate {b*1E4} [$10^{-4}$/turn]")
plt.title("PyHEADTAIL")
plt.legend(loc="upper left")
plt.xlabel("Turn")
plt.ylabel("x [$\sigma_x$]")

p_pht = particles

############ xsuite-PyHEADTAIL part (the WakeField instance is shared) ########################

particles = xt.Particles(
    _context=context,
    circumference=circumference,
    particlenumber_per_mp=bunch_intensity / n_macroparticles,
    q0=1,
    mass0=protonMass,
    gamma0=gamma,
    x=np.zeros(2*len(x0))
)

np2ctx = context.nparray_to_context_array
particles.x[::2] = np2ctx(x0)
particles.px[::2] = np2ctx(px0)
particles.y[::2] = np2ctx(y0)
particles.py[::2] = np2ctx(py0)
particles.zeta[::2] = np2ctx(zeta0)
particles.delta[::2] = np2ctx(delta0)
particles.state[1::2] = 0

#particles.x[:10000] = x0
#particles.px[:10000] = px0
#particles.y[:10000] = y0
#particles.py[:10000] = py0
#particles.zeta[:10000] = zeta0
#particles.delta[:10000] = delta0
#particles.state[10000:] = 0


arc = xt.LineSegmentMap(
    _context=context,
    betx=beta_x,
    bety=beta_y,
    qx=Q_x,
    qy=Q_y,
    bets=beta_s,
    qs=Q_s,
    dqx=chroma,
    det_xx=det_xx,
    det_xy=det_xy,
    det_yy=det_xx,
    det_yx=det_xy,
    energy_ref_increment=0.0,
    energy_increment=0,
)


line = xt.Line(elements=[arc, wake_field, damper],
                   element_names=['arc', 'wake_field', 'damper'])
line.build_tracker(_context=context)

t_xt_start = time.time()
turns = np.arange(nTurn)
x = np.zeros(nTurn, dtype=float)
for turn in range(nTurn):

    line.track(particles)

    x[turn] = np.average(particles.x[particles.state>0])
    if turn % 1000 == 0:
        print(
            f"PyHtXt - turn {turn}")
t_xt_end = time.time()

print(f'Xt full loop: {t_xt_end - t_xt_start}s')

x /= np.sqrt(normemit * beta_x / gamma / betar)
plt.figure(1)
plt.plot(turns, x, label=f"{i_oct}A")
iMin = 1000
iMax = nTurn - 1000
if iMin >= iMax:
    iMin = 0
    iMax = nTurn
ampl = np.abs(hilbert(x))
b, a, r, p, stderr = linregress(turns[iMin:iMax], np.log(ampl[iMin:iMax]))
gr_xtpyht = b
plt.plot(turns, np.exp(a + b * turns), "--k", label=f"{1/b:.3E} turns")
print(f"Growth rate {b*1E4} [$10^{-4}$/turn]")
plt.title("xsuite-PyHEADTAIL")
plt.legend(loc="upper left")
plt.xlabel("Turn")
plt.ylabel("x [$\sigma_x$]")

print(f'{gr_pyht=}, {gr_xtpyht=} {gr_pyht-gr_xtpyht=}')
plt.show()
assert np.isclose(gr_xtpyht, gr_pyht, rtol=1e-3, atol=1e-100)




"""
Beam Profile Monitor

Author: Philipp Niedermayer
Date: 2023-08-15
"""

import numpy as np

import xobjects as xo

from ..base_element import BeamElement
from ..beam_elements import Marker
from ..internal_record import RecordIndex
from ..general import _pkg_root

class BeamProfileMonitorRecord(xo.Struct):
    counts_x = xo.Float64[:]
    counts_y = xo.Float64[:]


class BeamProfileMonitor(BeamElement):

    _xofields={
        'particle_id_start': xo.Int64,
        'num_particles': xo.Int64,
        'start_at_turn': xo.Int64,
        'stop_at_turn': xo.Int64,
        'frev': xo.Float64,
        'sampling_frequency': xo.Float64,
        'nx': xo.Int64,
        'x_min': xo.Float64,
        'dx': xo.Float64,
        'ny': xo.Int64,
        'y_min': xo.Float64,
        'dy': xo.Float64,
        'sample_size': xo.Int64,
        '_index': RecordIndex,
        'data': BeamProfileMonitorRecord,
    }

    behaves_like_drift = True
    allow_loss_refinement = True

    properties = [field.name for field in BeamProfileMonitorRecord._fields]

    _extra_c_sources = [
        _pkg_root.joinpath('headers/atomicadd.h'),
        _pkg_root.joinpath('monitors/beam_profile_monitor.h')
    ]


    def __init__(self, *, particle_id_range=None, particle_id_start=None, num_particles=None,
                 start_at_turn=None, stop_at_turn=None, frev=None, sampling_frequency=None,
                 nx=None, x_range=None, ny=None, y_range=None,
                 n=None, range=None, _xobject=None, **kwargs):
        """
        Monitor to save the transverse profile of the tracked particles


        The monitor allows for arbitrary sampling rate and can thus not only be used to monitor
        bunch profiles, but also for coasting beams. Internally, the particle arrival time
        is used when determining the record index:

            i = sampling_frequency * ( ( at_turn - start_turn ) / f_rev - zeta / beta0 / c0 )

        where zeta=(s-beta0*c0*t) is the longitudinal coordinate of the particle, beta0 the
        relativistic beta factor of the particle, c0 is the speed of light, at_turn is the
        current turn number, f_rev is the revolution frequency, and sampling_frequency is the
        sampling frequency.

        Note that the index is rounded, i.e. the result array represents data of particles
        equally distributed around the reference particle. For example, if the sampling_frequency
        is twice the revolution frequency, the first item contains data from particles in the
        range zeta/circumference = -0.25 .. 0.25, the second item in the range 0.25 .. 0.75 and
        so on.


        The monitor provides the following data:
        - `x_intensity`, `y_intensity`: the profile intensity (particle count per bin) as 2D array of shape (size, n)
                                        where size = round(( stop_at_turn - start_at_turn ) * sampling_frequency / frev)
        - `x_edges`, `y_edges`: the profile edges (position) in m as 1D array of shape (n+1)
        - `x_grid`, `y_grid`: the profile position (midpoints) in m as 1D array of shape (n)

        Args:
            num_particles (int, optional): Number of particles to monitor. Defaults to -1 which means ALL.
            particle_id_start (int, optional): First particle id to monitor. Defaults to 0.
            particle_id_range (tuple, optional): Range of particle ids to monitor (start, stop). Stop is exclusive.
                                                 Defaults to (particle_id_start, particle_id_start+num_particles).
            start_at_turn (int): First turn of reference particle (inclusive) at which to monitor.
            stop_at_turn (int): Last turn of reference particle (exclusiv) at which to monitor.
            frev (float): Revolution frequency in Hz of circulating beam (used to relate turn number to sample index).
            sampling_frequency (float): Sampling frequency in Hz.
            nx (int, optional): Number of raster points of the horizontal profile. Defaults to 128.
            x_range (float or tuple): Extend of raster points of the profile in m. Either a tuple of (min_x, max_x)
                                             or a scalar `width` in which case a range of (-width/2, width/2) is used.
            ny (int, optional): Number of raster points of the vertical profile. Defaults to 128.
            y_range (float or tuple): Extend of raster points of the profile in m. Either a tuple of (min_y, max_y)
                                             or a scalar `width` in which case a range of (-width/2, width/2) is used.
            n: Default value for `nx` and `ny` if these are not set.
            range: Default value for `x_range` and `y_range` if these are not set.

        """
        if _xobject is not None:
            super().__init__(_xobject=_xobject)

        else:
            # dict parameters
            if particle_id_range is None:
                if particle_id_start is None:
                    particle_id_start = 0
                if num_particles is None:
                    num_particles = -1
            elif particle_id_start is None and num_particles is None:
                particle_id_start = particle_id_range[0]
                num_particles = particle_id_range[1] - particle_id_range[0]
            else:
                raise ValueError("Parameter `particle_id_range` must not be used together with `num_particles` and/or `particle_id_start`")
            if start_at_turn is None:
                start_at_turn = 0
            if stop_at_turn is None:
                stop_at_turn = 0
            if frev is None:
                frev = 1
            if sampling_frequency is None:
                sampling_frequency = 1

            if nx is None:
                nx = n or 128
            if x_range is None:
                if range is None:
                    raise ValueError("Either `x_range` or `range` must be provided")
                x_range = range
            if np.isscalar(x_range):
                x_range = (-x_range/2, x_range/2)
            x_min = x_range[0]
            dx = (x_range[1]-x_range[0])/nx

            if ny is None:
                ny = n or 128
            if y_range is None:
                if range is None:
                    raise ValueError("Either `y_range` or `range` must be provided")
                y_range = range
            if np.isscalar(y_range):
                y_range = (-y_range/2, y_range/2)
            y_min = y_range[0]
            dy = (y_range[1]-y_range[0])/ny

            sample_size = int(round(( stop_at_turn - start_at_turn ) * sampling_frequency / frev))

            if "data" not in kwargs:
                # explicitely init with zeros (instead of size only) to have consistent initial values
                kwargs["data"] = dict(
                    counts_x = np.zeros(sample_size*nx),
                    counts_y = np.zeros(sample_size*ny),
                )

            super().__init__(particle_id_start=particle_id_start, num_particles=num_particles,
                             start_at_turn=start_at_turn, stop_at_turn=stop_at_turn, frev=frev,
                             sampling_frequency=sampling_frequency, nx=nx,
                             x_min=x_min, dx=dx,
                             ny=ny, y_min=y_min,
                             dy=dy, sample_size=sample_size, **kwargs)


    def __repr__(self):
        return (
            f"{type(self).__qualname__}(start_at_turn={self.start_at_turn}, stop_at_turn={self.stop_at_turn}, "
            f"particle_id_start={self.particle_id_start}, num_particles={self.num_particles}, frev={self.frev}, "
            f"sampling_frequency={self.sampling_frequency}) at {hex(id(self))}"
        )

    @property
    def x_edges(self):
        """Edge positions of horizontal profile as array of shape (n+1)"""
        return self.x_min + self.dx * np.arange(self.nx + 1)

    @property
    def x_grid(self):
        """Profile positions (midpoints) of horizontal profile as array of shape (n)"""
        return self.x_edges[1:] - self.dx/2

    @property
    def x_intensity(self):
        """Horizontal profile intensity (particle count per bin) as 2D array of shape (size, n)
           where size = round(( stop_at_turn - start_at_turn ) * sampling_frequency / frev)"""
        x = self.data.counts_x.to_nparray()
        return x.reshape((-1, self.nx))

    @property
    def y_edges(self):
        """Edge positions of vertical profile as array of shape (n+1)"""
        return self.y_min + self.dy * np.arange(self.ny + 1)

    @property
    def y_grid(self):
        """Profile positions (midpoints) of vertical profile as array of shape (n)"""
        return self.y_edges[1:] - self.dy/2

    @property
    def y_intensity(self):
        """Vertical profile intensity (particle count per bin) as 2D array of shape (size, n)
           where size = round(( stop_at_turn - start_at_turn ) * sampling_frequency / frev)"""
        y = self.data.counts_y.to_nparray()
        return y.reshape((-1, self.ny))

    def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
        return Marker(_context=_context, _buffer=_buffer, _offset=_offset)



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import pymask as pm

# The parts marked by (*) in following need to be
# adapted according to knob definitions

def build_sequence(mad, beam):

    # Select beam
    mad.input(f'mylhcbeam = {beam}')

    # Make link to optics toolkit
    pm.make_links({'optics_toolkit': 'optics_repository/HLLHCV1.4/toolkit'},
        force=True)

    mad.input('''

        ! Specify machine version
        ver_lhc_run = 0;
        ver_hllhc_optics = 1.4;

        ! Get the toolkit
        call, file="optics_repository/HLLHCV1.4/toolkit/macro.madx";


        ! Build sequence
        option, -echo,-warn,-info;
        if (mylhcbeam==4){
          call,file="optics_repository/runIII/lhcb4.seq";
        } else {
          call,file="optics_repository/runIII/lhc.seq";
        };
        option, -echo, warn,-info;

        !Install HL-LHC
        call, file="optics_repository/HLLHCV1.4/hllhc_sequence.madx";


        ! Slice nominal sequence
        exec, myslice;

        ! Install placeholder elements for errors (set to zero)
        call, file="errors/HL-LHC/install_MQXF_fringenl.madx";    ! adding fringe place holder
        call, file="errors/HL-LHC/install_MCBXFAB_errors.madx";   ! adding D1 corrector placeholders in IR1/5 (for errors)
        call, file="errors/HL-LHC/install_MCBRD_errors.madx";     ! adding D2 corrector placeholders in IR1/5 (for errors)
        call, file="errors/HL-LHC/install_NLC_errors.madx";       ! adding non-linear corrector placeholders in IR1/5 (for errors)

        !Cycling w.r.t. to IP3 (mandatory to find closed orbit in collision in the presence of errors)
        if (mylhcbeam<3){
          seqedit, sequence=lhcb1; flatten; cycle, start=IP3; flatten; endedit;
        };
        seqedit, sequence=lhcb2; flatten; cycle, start=IP3; flatten; endedit;
        ''')

def apply_optics(mad, optics_file):
    mad.call(optics_file)


def set_optics_specific_knobs(mad, knob_settings, mode=None):

    # Copy knob settings to mad variable space
    mad.set_variables_from_dict(params=knob_settings)

    # A check
    if mad.globals.nrj < 500:
        assert knob_settings['on_disp'] == 0

    # A knob redefinition
    mad.input('on_alice := on_alice_normalized * 7000./nrj;')
    mad.input('on_lhcb := on_lhcb_normalized * 7000./nrj;')


def twiss_and_check(mad, sequences_to_check, twiss_fname,
        tol_beta=1e-3, tol_sep=1e-6, save_twiss_files=True,
        check_betas_at_ips=True, check_separations_at_ips=True):

    var_dict = mad.get_variables_dicts()
    twiss_dfs = {}
    summ_dfs = {}
    for ss in sequences_to_check:
        mad.use(ss)
        mad.twiss()
        tdf = mad.get_twiss_df('twiss')
        twiss_dfs[ss] = tdf
        sdf = mad.get_summ_df('summ')
        summ_dfs[ss] = sdf

    if save_twiss_files:
        for ss in sequences_to_check:
            tt = twiss_dfs[ss]
            if twiss_fname is not None:
                tt.to_parquet(twiss_fname + f'_seq_{ss}.parquet')

    if check_betas_at_ips:
        for ss in sequences_to_check:
            tt = twiss_dfs[ss]
            _check_beta_at_ips_against_madvars(beam=ss[-1],
                    twiss_df=tt,
                    variable_dicts=var_dict,
                    tol=tol_beta)
        print('IP beta test against knobs passed!')

    if check_separations_at_ips:
        twiss_df_b1 = twiss_dfs['lhcb1']
        twiss_df_b2 = twiss_dfs['lhcb2']
        _check_separations_at_ips_against_madvars(twiss_df_b1, twiss_df_b2,
                var_dict, tol=tol_sep)
        print('IP separation test against knobs passed!')

    other_data = {}
    other_data.update(var_dict)
    other_data['summ_dfs'] = summ_dfs

    return twiss_dfs, other_data

def lumi_control(mad, twiss_dfs, configuration, knob_names):
    from scipy.optimize import least_squares

    # Leveling in IP8
    sep_plane_ip8 = configuration['sep_plane_ip8']
    sep_knobname_ip8 = knob_names['sepknob_ip8_mm']

    L_target_ip8 = configuration['lumi_ip8']
    def function_to_minimize_ip8(sep8_m):
        my_dict_IP8=pm.get_luminosity_dict(
            mad, twiss_dfs, 'ip8', configuration['nco_IP8'])
        my_dict_IP8[sep_plane_ip8 + '_1']=np.abs(sep8_m)
        my_dict_IP8[sep_plane_ip8 + '_2']=-np.abs(sep8_m)
        return np.abs(pm.luminosity(**my_dict_IP8) - L_target_ip8)
    sigma_sep_b1_ip8=np.sqrt(twiss_dfs['lhcb1'].loc['ip8:1']['bet'+sep_plane_ip8]
               * mad.sequence.lhcb1.beam['e'+sep_plane_ip8])
    optres_ip8=least_squares(function_to_minimize_ip8, sigma_sep_b1_ip8)
    mad.globals[sep_knobname_ip8] = (np.sign(mad.globals[sep_knobname_ip8])
                                * np.abs(optres_ip8['x'][0])*1e3)

    # Halo collision in IP2
    sep_plane_ip2 = configuration['sep_plane_ip2']
    sep_knobname_ip2 = knob_names['sepknob_ip2_mm']
    sigma_sep_b1_ip2=np.sqrt(
            twiss_dfs['lhcb1'].loc['ip2:1']['bet'+sep_plane_ip2]
            * mad.sequence.lhcb1.beam['e'+sep_plane_ip2])
    mad.globals[sep_knobname_ip2] = (np.sign(mad.globals[sep_knobname_ip2])
            * configuration['fullsep_in_sigmas_ip2']*sigma_sep_b1_ip2/2*1e3)


def _check_beta_at_ips_against_madvars(beam, twiss_df, variable_dicts, tol):
    twiss_value_checks=[]
    for iip, ip in enumerate([1,2,5,8]):
        for plane in ['x', 'y']:
            # (*) Adapet based on knob definitions
            twiss_value_checks.append({
                    'element_name': f'ip{ip}:1',
                    'keyword': f'bet{plane}',
                    'varname': f'bet{plane}ip{ip}b{beam}',
                    'tol': tol[iip]})

    pm.check_twiss_against_madvars(twiss_value_checks, twiss_df, variable_dicts)

def _check_separations_at_ips_against_madvars(twiss_df_b1, twiss_df_b2,
        variables_dict, tol):

    separations_to_check = []
    for iip, ip in enumerate([1,2,5,8]):
        for plane in ['x', 'y']:
            # (*) Adapet based on knob definitions
            separations_to_check.append({
                    'element_name': f'ip{ip}:1',
                    'scale_factor': -2*1e-3,
                    'plane': plane,
                    # knobs like on_sep1h, onsep8v etc
                    'varname': f'on_sep{ip}'+{'x':'h', 'y':'v'}[plane],
                    'tol': tol[iip]})
    pm.check_separations_against_madvars(separations_to_check,
            twiss_df_b1, twiss_df_b2, variables_dict)



import numpy as np
from cpymad.madx import Madx

import xtrack as xt

mad = Madx()
mad.call('../../test_data/hllhc15_noerrors_nobb/sequence_with_crabs.madx')
mad.use('lhcb1')
mad.globals.on_crab1 = -190
mad.globals.on_crab5 = -190

line = xt.Line.from_madx_sequence(mad.sequence.lhcb1)
line.particle_ref = xt.Particles(p0c=7000e9, mass0=xt.PROTON_MASS_EV)

line.build_tracker()

# Measure crabbing angle at IP1 and IP5
z1 = 1e-4
z2 = -1e-4

tw1 = line.twiss(zeta0=z1).to_pandas()
tw2 = line.twiss(zeta0=z2).to_pandas()

tw1.set_index('name', inplace=True)
tw2.set_index('name', inplace=True)

phi_c_ip1 = ((tw1.loc['ip1', 'x'] - tw2.loc['ip1', 'x'])
         /(tw1.loc['ip1', 'zeta'] - tw2.loc['ip1', 'zeta']))

phi_c_ip5 = ((tw1.loc['ip5', 'y'] - tw2.loc['ip5', 'y'])
            /(tw1.loc['ip5', 'zeta'] - tw2.loc['ip5', 'zeta']))

assert np.isclose(phi_c_ip1, -190e-6, atol=1e-7, rtol=0)
assert np.isclose(phi_c_ip5, -190e-6, atol=1e-7, rtol=0)



import os
import time

from cpymad.madx import Madx

from cpymad.madx import Madx
mad=Madx()
mad.call('../../../hllhc15/lhc/lhc.seq')
mad.call('../../../hllhc15/hllhc_sequence.madx')
mad.input('beam, sequence=lhcb1, particle=proton, energy=7000;')
mad.input('beam, sequence=lhcb2, particle=proton, energy=7000, bv=-1;')
mad.use('lhcb1')
mad.call("../../../hllhc15/round/opt_round_150_1500.madx")
mad.twiss()

mad.call("../../../hllhc15/toolkit/macro.madx")

# Get the absolute path of the current file
current_dir = os.path.dirname(os.path.abspath(__file__))
link_path = os.path.join(current_dir, 'slhc')
target_path = os.path.join(current_dir, '../../../hllhc15/')

# Create symlink if it doesn't exist
if not os.path.exists(link_path):
    os.symlink(target_path, link_path)

mad.input('''
    mux_ip15=30.830335;
    muy_ip15=29.869337;
''')

t1 = time.time()
mad.input('''
weakdeltaphasearc(BIM): macro = {
	exec,phasearc(2,3,23,BIM);
	exec,phasearc(3,4,34,BIM);
	exec,phasearc(6,7,67,BIM);
	exec,phasearc(7,8,78,BIM);
	dmux15=mux23BIM+mux34BIM;
	dmuy15=muy23BIM+muy34BIM;
	dmux51=mux67BIM+mux78BIM;
	dmuy51=muy67BIM+muy78BIM;
};

mkdelta_ip15(BIM,tmux,tmuy): macro = {
	exec,weakdeltaphasearc(BIM);
	dmuxtotaux15=dmux15; dmuytotaux15=dmuy15;
	dmuxtotaux51=dmux51; dmuytotaux51=dmuy51;
	match,use_macro;
	vary, name=kqf.a34;
	vary, name=kqf.a78;
	vary, name=kqd.a34;
	vary, name=kqd.a78;
	vary, name=kqf.a23;
	vary, name=kqf.a67;
	vary, name=kqd.a23;
	vary, name=kqd.a67;
	use_macro,name=weakdeltaphasearc(BIM);
	constraint,expr=  dmux15=dmuxtotaux15+tmux;
	constraint,expr=  dmuy15=dmuytotaux15+tmuy;
	constraint,expr=  dmux51=dmuxtotaux51-tmux;
	constraint,expr=  dmuy51=dmuytotaux51-tmuy;
	jacobian,calls=10,tolerance=1.0e-19;
	endmatch;
	!kqf.a23=kqf.a23; kqf.a34=kqf.a34; kqf.a67=kqf.a67; kqf.a78=kqf.a78;
	!kqd.a23=kqd.a23; kqd.a34=kqd.a34; kqd.a67=kqd.a67; kqd.a78=kqd.a78;
};
''')


mad.input('''
	exec,check_ip(b1);
	dphix15=table(twiss,IP1,mux)-table(twiss,IP5,mux);
	dphiy15=table(twiss,IP1,muy)-table(twiss,IP5,muy);
	if (dphix15<0){ dphix15=dphix15+refqxb1;};
	if (dphiy15<0){ dphiy15=dphiy15+refqyb1;};
	value,dphix15,dphiy15;
	dphix15= dphix15-mux_ip15;
	dphiy15= dphiy15-muy_ip15;
	value,dphix15,dphiy15;
	exec,mkdelta_ip15(b1,dphix15,dphiy15);
''')
t2 = time.time()

mad.input('''

	! select all insertion
	jac_calls=15;
	jac_tol=1e-20;
	exec,select(7,67,78,b1);
	exec,select(3,23,34,b1);
	scxir1=betx_IP1/betx0_IP1; scyir1=bety_IP1/bety0_IP1;
	scxir5=betx_IP5/betx0_IP5; scyir5=bety_IP5/bety0_IP5;
	value,scxir1,scyir1,scxir5,scyir5;
	exec,selectIRAUX(7,8,1,2,3,b1,scxir1,scyir1,betx0_IP1,bety0_IP1);
	exec,selectIRAUX(3,4,5,6,7,b1,scxir5,scyir5,betx0_IP5,bety0_IP5);


	! rematch all insertion
	relax_match15=0;
	call,file="../../../hllhc15/toolkit/rematch_ir234678_b1.madx";
	value,tarir2b1,tarir3b1,tarir4b1,tarir6b1,tarir7b1,tarir8b1;
	value,relax_match15;
	while (tarir3b1+tarir7b1+tarir4b1+tarir6b1+tarir2b1+tarir8b1>1e-10){
		if (relax_match15>=3){
	    	print, text="Error in rematching phase change";
	    	tarir2b1=0; tarir3b1=0; tarir4b1=0; tarir6b1=0; tarir7b1=0; tarir8b1=0;
	    	exit;
		};
		if (relax_match15<3){
	    	relax_match15=relax_match15+1;
	    	call,file="../../../hllhc15/toolkit/rematch_ir234678_b1.madx";
	    };
	};

''')
t3 = time.time()

print('\nTime match arcs: ', t2-t1)
print('\nTime match IRs: ', t3-t2)


import time
import numpy as np
import xtrack as xt

line = xt.Line.from_json(
    'LHCB1_qx62.275_qy60.293_qp10_mo0.0_cmr0.000_phasechange0.0.json')
line.cycle('ip1', inplace=True)

line.build_tracker()

line.twiss_default['method'] = '4d'

line.vars['vrf400'] = 0.0

tw = line.twiss()

tw_mo = tw.rows['mo.*.b1']

def get_rdt(tmux, tmuy, betx, bety, mux, muy, j=2, k=0, l=0, m=2, q1=62.27, q2=60.295):
    dphix = mux-tmux
    dphix[np.where(dphix <= 0 )] += q1
    dphiy = muy-tmuy
    dphiy[np.where(dphiy <= 0 )] += q2
    phasor = np.exp(np.pi*2.j*((j-k)*(dphix) + (l-m)*(dphiy)))
    amp = np.sqrt(betx)**(j+k)*np.sqrt(bety)**(l+m)
    return np.sum(amp*phasor)#/(1-np.exp(2*np.pi*1.j*((j-k)*q1 + (l-m)*q2)))

rdt = get_rdt(tmux=0, tmuy=0, betx=tw_mo['betx'], bety=tw_mo['bety'],
        mux=tw_mo['mux'], muy=tw_mo['muy'], q1=tw.qx, q2=tw.qy)

class RDTarget(xt.Target):
    def __init__(self, j=2, k=0, l=0, m=2):
        self.j = j
        self.k = k
        self.l = l
        self.m = m

        xt.Target.__init__(
            self, tag='rdt', tar=self.compute_abs_target, value=0, tol=500, weight=1e-6)

    def compute_target(self, tw):
        tw_mo = tw.rows['mo.*.b1']
        rdt = get_rdt(tmux=0, tmuy=0, betx=tw_mo['betx'], bety=tw_mo['bety'],
            mux=tw_mo['mux'], muy=tw_mo['muy'], q1=tw.mux[-1], q2=tw.muy[-1],
            j=self.j, k=self.k, l=self.l, m=self.m)
        return rdt

    # def __repr__(self):
    #     return f'RDTarget(j={self.j}, k={self.k}, l={self.l}, m={self.m})'

class RDTargetReal(RDTarget):
    def compute_abs_target(self, tw):
        return super().compute_target(tw).real

class RDTargetImag(RDTarget):
    def compute_abs_target(self, tw):
        return super().compute_target(tw).imag

lims = {}
varlist = ['kqtf.a12b1', 'kqtf.a23b1', 'kqtf.a34b1', 'kqtf.a45b1',
           'kqtf.a56b1', 'kqtf.a67b1', 'kqtf.a78b1', 'kqtf.a81b1',
           'kqtd.a12b1', 'kqtd.a23b1', 'kqtd.a34b1', 'kqtd.a45b1',
           'kqtd.a56b1', 'kqtd.a67b1', 'kqtd.a78b1', 'kqtd.a81b1']

for var in varlist:
    val = line.vars[var]._value
    #vals = [ 0.8*val, 1.2*val]
    #offset = 0.0001
    lims[var] = ( -0.001,  +0.001)
    #lims[var] = ( -offset + min(vals),  offset + max(vals))


tw0 = line.twiss()

opt = line.match_knob(
    knob_name='phase_knob',
    run=False,
    assert_within_tol=False,
    init=tw.get_twiss_init(0),
    start=0, end=len(line)-1,
    targets=[
        RDTargetReal(j=2, k=0, l=0, m=2),
        RDTargetReal(j=4, k=0, l=0, m=0),
        # RDTargetReal(j=0, k=0, l=4, m=0),
        RDTargetImag(j=2, k=0, l=0, m=2),
        RDTargetImag(j=4, k=0, l=0, m=0),
        # RDTargetImag(j=0, k=0, l=4, m=0),
        xt.TargetSet(tag="tune", tars=['mux', 'muy'], value=tw0, at='lhcb1$end')
    ],
    vary=[
        xt.Vary('kqtf.a12b1', step=1.e-8, limits=lims['kqtf.a12b1']),
        xt.Vary('kqtf.a23b1', step=1.e-8, limits=lims['kqtf.a23b1']),
        xt.Vary('kqtf.a34b1', step=1.e-8, limits=lims['kqtf.a34b1']),
        xt.Vary('kqtf.a45b1', step=1.e-8, limits=lims['kqtf.a45b1']),
        xt.Vary('kqtf.a56b1', step=1.e-8, limits=lims['kqtf.a56b1']),
        xt.Vary('kqtf.a67b1', step=1.e-8, limits=lims['kqtf.a67b1']),
        xt.Vary('kqtf.a78b1', step=1.e-8, limits=lims['kqtf.a78b1']),
        xt.Vary('kqtf.a81b1', step=1.e-8, limits=lims['kqtf.a81b1']),
        xt.Vary('kqtd.a12b1', step=1.e-8, limits=lims['kqtd.a12b1']),
        xt.Vary('kqtd.a23b1', step=1.e-8, limits=lims['kqtd.a23b1']),
        xt.Vary('kqtd.a34b1', step=1.e-8, limits=lims['kqtd.a34b1']),
        xt.Vary('kqtd.a45b1', step=1.e-8, limits=lims['kqtd.a45b1']),
        xt.Vary('kqtd.a56b1', step=1.e-8, limits=lims['kqtd.a56b1']),
        xt.Vary('kqtd.a67b1', step=1.e-8, limits=lims['kqtd.a67b1']),
        xt.Vary('kqtd.a78b1', step=1.e-8, limits=lims['kqtd.a78b1']),
        xt.Vary('kqtd.a81b1', step=1.e-8, limits=lims['kqtd.a81b1']),
        # xt.VaryList(
        #     ['kqtf.a12b1', 'kqtf.a23b1', 'kqtf.a34b1', 'kqtf.a45b1',
        #     'kqtf.a56b1', 'kqtf.a67b1', 'kqtf.a78b1', 'kqtf.a81b1'], step=1e-6),
        # xt.VaryList(
        #     ['kqtd.a12b1', 'kqtd.a23b1', 'kqtd.a34b1', 'kqtd.a45b1',
        #     'kqtd.a56b1', 'kqtd.a67b1', 'kqtd.a78b1', 'kqtd.a81b1'], step=1e-6),
    ]
)

t1 = time.time()
opt.step(30)

opt.disable_targets(tag='rdt')
opt.solve()
t2 = time.time()
print(f'Optimization time: {t2-t1:.2f} s')

knob_values_xs = opt.get_knob_values()
target_values_xs = opt.target_status(ret=True).current_val
tw_xs = line.twiss()

# for ii in range(len(opt.vary)):
#     opt.vary[ii].limits = [opt.vary[ii].limits[0] - offset, opt.vary[ii].limits[1] + offset]

knob_values_ng  = {
    'kqtf.a12b1_from_phase_knob': -0.0022477200000,
    'kqtf.a23b1_from_phase_knob': -0.0006109026670,
    'kqtf.a34b1_from_phase_knob': -0.0006740726670,
    'kqtf.a45b1_from_phase_knob': +0.0015222900000,
    'kqtf.a56b1_from_phase_knob': +0.0011189300000,
    'kqtf.a67b1_from_phase_knob': +0.0020387763940,
    'kqtf.a78b1_from_phase_knob': -0.0011010306070,
    'kqtf.a81b1_from_phase_knob': -0.0001300250000,
    'kqtd.a12b1_from_phase_knob': -0.0001437190000,
    'kqtd.a23b1_from_phase_knob': +0.0010619748420,
    'kqtd.a34b1_from_phase_knob': +0.0001529048423,
    'kqtd.a45b1_from_phase_knob': -0.0004891330000,
    'kqtd.a56b1_from_phase_knob': +0.0008419600000,
    'kqtd.a67b1_from_phase_knob': +0.0016072722540,
    'kqtd.a78b1_from_phase_knob': -0.0013696167460,
    'kqtd.a81b1_from_phase_knob': -0.0016425400000,
}

line.vars.update(knob_values_ng)
target_values_ng = opt.target_status(ret=True).current_val
tw_ng = line.twiss()

print(f'xs: |f2002| = \t{np.sqrt(target_values_xs[0]**2 + target_values_xs[1]**2):.2e} '
        f'\t|f4000| = {np.sqrt(target_values_xs[2]**2 + target_values_xs[3]**2):.2e}')
print(f'ng: |f2002| = \t{np.sqrt(target_values_ng[0]**2 + target_values_ng[1]**2):.2e} '
        f'\t|f4000| = {np.sqrt(target_values_ng[2]**2 + target_values_ng[3]**2):.2e}')

import matplotlib.pyplot as plt
plt.figure(1)
sp1 = plt.subplot(211)
plt.plot(tw.s, tw_ng.betx/tw.betx-1, label='ng')
plt.plot(tw.s, tw_xs.betx/tw.betx-1, label='xs')
plt.ylabel(r'$\Delta \beta_x / \beta_x$')
plt.legend()
plt.subplot(212, sharex=sp1)
plt.plot(tw.s, tw_ng.bety/tw.bety-1, label='ng')
plt.plot(tw.s, tw_xs.bety/tw.bety-1, label='xs')
plt.ylabel(r'$\Delta \beta_y / \beta_y$')
plt.xlabel('s [m]')
plt.legend()
plt.show()



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
import xtrack as xt

# Load a line and build tracker
line = xt.Line.from_json(
    '../../test_data/hllhc15_noerrors_nobb/line_and_particle.json')
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, q0=1, energy0=7e12)
line.build_tracker()

# Twiss
tw = line.twiss()

# Transverse normalized emittances
nemitt_x = 2.5e-6
nemitt_y = 2.5e-6

# Longitudinal emittance from energy spread
sigma_pzeta = 2e-4
gemitt_zeta = sigma_pzeta**2 * tw.bets0
# similarly, if the bunch length is known, the emittance can be computed as
# gemitt_zeta = sigma_zeta**2 / tw.bets0

# Compute beam sizes
beam_sizes = tw.get_beam_covariance(nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                                    gemitt_zeta=gemitt_zeta)

# Inspect beam sizes (table can be accessed similarly to twiss tables)
beam_sizes.rows['ip.?'].show()
# prints
#
# name                       s     sigma_x     sigma_y sigma_zeta    sigma_px ...
# ip3                        0 0.000226516 0.000270642    0.19694 4.35287e-06
# ip4                  3332.28 0.000281326 0.000320321   0.196941 1.30435e-06
# ip5                  6664.57  7.0898e-06 7.08975e-06    0.19694  4.7265e-05
# ip6                  9997.01 0.000314392 0.000248136   0.196939 1.61401e-06
# ip7                  13329.4 0.000205156 0.000223772   0.196939 2.70123e-06
# ip8                  16650.7 2.24199e-05 2.24198e-05   0.196939 1.49465e-05
# ip1                  19994.2 7.08975e-06 7.08979e-06   0.196939 4.72651e-05
# ip2                  23326.6 5.78877e-05 5.78878e-05   0.196939 5.78879e-06

# All covariances are computed including those from linear coupling
beam_sizes.keys()
# is:
#
# ['s', 'name', 'sigma_x', 'sigma_y', 'sigma_zeta', 'sigma_px', 'sigma_py',
# 'sigma_pzeta', 'Sigma', 'Sigma11', 'Sigma12', 'Sigma13', 'Sigma14', 'Sigma15',
# 'Sigma16', 'Sigma21', 'Sigma22', 'Sigma23', 'Sigma24', 'Sigma25', 'Sigma26',
# 'Sigma31', 'Sigma32', 'Sigma33', 'Sigma34', 'Sigma41', 'Sigma42', 'Sigma43',
# 'Sigma44', 'Sigma51', 'Sigma52'])

# Plot
import matplotlib.pyplot as plt
plt.close('all')

fig1 = plt.figure(1, figsize=(6.4, 4.8*1.5))
spbet = plt.subplot(3,1,1)
spdisp = plt.subplot(3,1,2, sharex=spbet)
spbsz = plt.subplot(3,1,3, sharex=spbet)

spbet.plot(tw.s, tw.betx)
spbet.plot(tw.s, tw.bety)
spbet.set_ylabel(r'$\beta_{x,y}$ [m]')

spdisp.plot(tw.s, tw.dx)
spdisp.plot(tw.s, tw.dy)
spdisp.set_ylabel(r'$D_{x,y}$ [m]')

spbsz.plot(beam_sizes.s, beam_sizes.sigma_x)
spbsz.plot(beam_sizes.s, beam_sizes.sigma_y)
spbsz.set_ylabel(r'$\sigma_{x,y}$ [m]')
spbsz.set_xlabel('s [m]')

spbet.set_xlim(tw['s', 'ip5'] - 2000, tw['s', 'ip5'] + 2000)

fig1.subplots_adjust(left=.15, right=.92, hspace=.27)
plt.show()

#!end-doc-part
assert np.allclose(beam_sizes.sigma_pzeta, 2e-4, atol=0, rtol=2e-5)
assert np.allclose(
    beam_sizes.sigma_zeta / beam_sizes.sigma_pzeta, tw.bets0, atol=0, rtol=5e-5)



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
import xtrack as xt

q1 = xt.Multipole(knl=[0, 2e-2])
q2 = xt.Multipole(knl=[0, -2e-2])

line = xt.Line(
        elements=[
        xt.Drift(length=1),
        xt.Cavity(frequency=400e6, voltage=1e6, lag=0),
        q1,
        xt.Drift(length=1),
        xt.Multipole(knl=[1e-2], hxl=1e-2),
        q2,
        ])

line.build_tracker()
line.config.XTRACK_GLOBAL_XY_LIMIT = 1e10

line.particle_ref = xt.Particles(p0c=6500e9)
import matplotlib.pyplot as plt
plt.close('all')
plt.figure()
ax1 = plt.subplot(211)
ax2 = plt.subplot(212, sharex=ax1)
fig2 = plt.figure()
axe = plt.subplot()
eigs = []
factors = np.linspace(80, 120, 100)
for factor in factors:
    q1.knl[1] = factor*0.020
    q2.knl[1] = -factor*0.020
    line.track(xt.Particles(p0c=6500e9, x=0.01, y=0.01),
                  num_turns=5000, turn_by_turn_monitor=True)
    mon = line.record_last_track
    RR = line.compute_one_turn_matrix_finite_differences(
                                               particle_on_co=line.particle_ref)
    ax1.semilogy(np.abs(mon.x.T), label=f'{factor}, {np.trace(RR["R_matrix"])}')
    ax2.semilogy(np.abs(mon.y.T), label=f'{factor}, {np.trace(RR["R_matrix"])}')

    w0, v0 = np.linalg.eig(xt.linear_normal_form.healy_symplectify(RR['R_matrix']))
    eigs.append(w0)
    #axe.plot(w0.real, w0.imag, '.')

eigs = np.array(eigs)

plt.legend(loc='upper left')

plt.figure()
for ii in range(6):
    plt.plot(factors, np.abs(eigs[:, ii]).T, '.')
plt.show()



import xtrack as xt

env = xt.Environment()

env.vars.default_to_zero = True
line = env.new_line(components=[
    env.new('b1', xt.Bend, length=1, knl=['a', 'b', 'c'], ksl=['d', 'e', 'f']),
    env.new('q1', xt.Quadrupole, length=1, knl=['a', 'b', 'c'], ksl=['d', 'e', 'f']),
    env.new('s1', xt.Sextupole, length=1, knl=['a', 'b', 'c'], ksl=['d', 'e', 'f']),
    env.new('o1', xt.Octupole, length=1, knl=['a', 'b', 'c'], ksl=['d', 'e', 'f']),
    env.new('s2', xt.Solenoid, length=1, knl=['a', 'b', 'c'], ksl=['d', 'e', 'f']),
    env.new('m1', xt.Multipole, length=1, knl=['a', 'b', 'c'], ksl=['d', 'e', 'f']),
])

# Set the strengths
env['a'] = 1
env['b'] = 2
env['c'] = 3
env['d'] = 4
env['e'] = 5
env['f'] = 6

# Extend knl and ksl for selected elements
line.extend_knl_ksl(order=10, element_names=['b1', 'q1'])

env['b1'].knl # is [1., 2., 3., 0., 0., 0., 0., 0., 0., 0., 0.]

# Extend knl and ksl for all elements
line.extend_knl_ksl(order=10)

env['s2'].knl # is [1., 2., 3., 0., 0., 0., 0., 0., 0., 0., 0.]



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import xtrack as xt

# Load a line and build tracker
line = xt.Line.from_json(
    '../../test_data/hllhc15_noerrors_nobb/line_and_particle.json')
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, q0=1, energy0=7e12)
line.build_tracker()

# Twiss (built-in defaults)
tw_a = line.twiss()
tw_a.method # is '6d'
tw_a.reference_frame # is 'proper'

# Inspect twiss defaults
line.twiss_default # is {}

# Set some twiss defaults
line.twiss_default['method'] = '4d'
line.twiss_default['reverse'] = True

# Twiss (defaults redefined)
tw_b = line.twiss()
tw_b.method # is '4d'
tw_b.reference_frame # is 'reverse'

# Inspect twiss defaults
line.twiss_default # is {'method': '4d', 'reverse': True}

# Reset twiss defaults
line.twiss_default.clear()

# Twiss (defaults reset)
tw_c = line.twiss()
tw_c.method # is '6d'
tw_c.reference_frame # is 'proper'



import xtrack as xt
import xdeps as xd

container_type = 'line'

env = xt.Environment()
line = xt.Line()

if container_type == 'env':
    ee = env
    eenv = env
elif container_type == 'line':
    ee = line
    eenv = line.env

ee['a']  = 3.
ee['b1']  = 3 * ee['a'] # done by value
ee['b2']  = 3 * ee.ref['a'] # done by reference
ee['c']  = '4 * a'

assert isinstance(ee['a'], float)
assert isinstance(ee['b1'], float)
assert isinstance(ee['b2'], float)
assert isinstance(ee['c'], float)

assert ee['a'] == 3
assert ee['b1'] == 9
assert ee['b2'] == 9
assert ee['c'] == 12

assert ee.ref['a']._value == 3
assert ee.ref['b1']._value == 9
assert ee.ref['b2']._value == 9
assert ee.ref['c']._value == 12

assert ee.get('a') == 3
assert ee.get('b1') == 9
assert ee.get('b2') == 9
assert ee.get('c') == 12

eenv.new('mb', 'Bend', extra={'description': 'Hello Riccarco'},
        k1='3*a', h=4*ee.ref['a'], knl=[0, '5*a', 6*ee.ref['a']])
assert isinstance(ee['mb'].k1, float)
assert isinstance(ee['mb'].h, float)
assert isinstance(ee['mb'].knl[0], float)
assert ee['mb'].k1 == 9
assert ee['mb'].h == 12
assert ee['mb'].knl[0] == 0
assert ee['mb'].knl[1] == 15
assert ee['mb'].knl[2] == 18

ee['a'] = 4
assert ee['a'] == 4
assert ee['b1'] == 9
assert ee['b2'] == 12
assert ee['c'] == 16
assert ee['mb'].k1 == 12
assert ee['mb'].h == 16
assert ee['mb'].knl[0] == 0
assert ee['mb'].knl[1] == 20
assert ee['mb'].knl[2] == 24

ee['mb'].k1 = '30*a'
ee['mb'].h = 40 * ee.ref['a']
ee['mb'].knl[1] = '50*a'
ee['mb'].knl[2] = 60 * ee.ref['a']
assert ee['mb'].k1 == 120
assert ee['mb'].h == 160
assert ee['mb'].knl[0] == 0
assert ee['mb'].knl[1] == 200
assert ee['mb'].knl[2] == 240

assert isinstance(ee['mb'].k1, float)
assert isinstance(ee['mb'].h, float)
assert isinstance(ee['mb'].knl[0], float)

assert ee.ref['mb'].k1._value == 120
assert ee.ref['mb'].h._value == 160
assert ee.ref['mb'].knl[0]._value == 0
assert ee.ref['mb'].knl[1]._value == 200
assert ee.ref['mb'].knl[2]._value == 240

assert ee.get('mb').k1 == 120
assert ee.get('mb').h == 160
assert ee.get('mb').knl[0] == 0
assert ee.get('mb').knl[1] == 200
assert ee.get('mb').knl[2] == 240

# Some interesting behavior
assert type(ee['mb']) is xd.madxutils.View
assert ee['mb'].__class__ is xt.Bend
assert type(ee.ref['mb']._value) is xt.Bend
assert type(ee.get('mb')) is xt.Bend



import pathlib

import numpy as np
from cpymad.madx import Madx

import xobjects as xo
import xpart as xp
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts

test_data_folder = pathlib.Path(
    __file__).parent.joinpath('../test_data').absolute()

@for_all_test_contexts
def test_twiss_psb(test_context):

    mad = Madx(stdout=False)
    mad.call(str(test_data_folder / 'psb_injection/psb_injection.seq'))
    mad.use('psb')
    twmad = mad.twiss()

    line = xt.Line.from_madx_sequence(mad.sequence['psb'])
    line.particle_ref = xp.Particles(mass0=xp.PROTON_MASS_EV, q0=1.,
                                    gamma0=mad.sequence['psb'].beam.gamma)

    line.build_tracker(_context=test_context)
    beta0 = line.particle_ref.beta0[0]

    tw = line.twiss()

    # With the approximation beta ~= beta0 we have delta ~= pzeta ~= 1/beta0 ptau
    # ==> ptau ~= beta0 delta ==> dptau / ddelta~= beta0

    xo.assert_allclose(tw.dqx, twmad.summary.dq1 * beta0, rtol=0, atol=1e-5)
    xo.assert_allclose(tw.dqy, twmad.summary.dq2 * beta0, rtol=0, atol=1e-5)

    dx_ref = np.interp(tw.s, twmad.s, twmad.dx * beta0)
    betx_ref = np.interp(tw.s, twmad.s, twmad.betx)
    bety_ref = np.interp(tw.s, twmad.s, twmad.bety)

    xo.assert_allclose(tw.dx, dx_ref, rtol=0, atol=1e-3)
    xo.assert_allclose(tw.betx, betx_ref, rtol=1e-5, atol=0)
    xo.assert_allclose(tw.bety, bety_ref, rtol=1e-5, atol=0)

    xo.assert_allclose(tw.momentum_compaction_factor, twmad.summary.alfa,
                      atol=0, rtol=5e-3)

